@@ -6,7 +6,6 @@
 #undef HAVE_STRUCT_OSOCKADDR
 #undef HAVE_LIBSM
 #undef HAVE_LIBINTL_H
-#undef HAVE_OBSTACK
 #undef HAVE_PROGRAM_INVOCATION_NAME
 #undef HAVE_PROGRAM_INVOCATION_SHORT_NAME
 #undef HAVE_SIGACTION_H
@@ -15,6 +14,8 @@
 #undef BALSA_PATCHLEVEL
 #undef BALSA_REVISION
 #undef BALSA_VERSION
+#undef BALSA_SHOW_ALL
+#undef BALSA_SHOW_INFO
 #undef PACKAGE
 #undef VERSION
 
@@ -25,3 +26,16 @@
 #undef SENDMAIL
 #undef USE_POP
 #undef USE_IMAP
+
+/* Shall we multithread? */
+#undef BALSA_USE_THREADS
+/* Try new init druid based on GOB */
+#undef BALSA_USE_EXPERIMENTAL_INIT
+/* Use GOB */
+#undef BALSA_HAVE_GOB
+/* We have root privs */
+#undef BALSA_SYSTEM_INSTALL
+/* We don't have root or won't use it */
+#undef BALSA_LOCAL_INSTALL
+/* Our data directory prefix */
+#undef BALSA_RESOURCE_PREFIX
