@@ -1,5 +1,5 @@
 /* Balsa E-Mail Client
- * Copyright (C) 1997-98 Jay Painter and Stuart Parmenter
+ * Copyright (C) 1997-1999 Jay Painter and Stuart Parmenter
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,41 +21,46 @@
 
 #include <string.h>
 #include <gnome.h>
-#include "balsa-app.h"
 
-void 
+void mutt_message (const char *fmt,...);
+void mutt_exit (int code);
+int mutt_yesorno (const char *msg, int def);
+int mutt_any_key_to_continue (const char *s);
+void mutt_clear_error (void);
+
+void
 mutt_message (const char *fmt,...)
 {
-  if (balsa_app.debug)
-    {
-      va_list ap;
-      char outstr[522];
+#ifdef DEBUG
+  va_list ap;
+  char outstr[522];
 
-      va_start (ap, fmt);
-      vsprintf (outstr, fmt, ap);
-      va_end (ap);
-      g_print ("mutt_message: %s\n", outstr);
-    }
+  va_start (ap, fmt);
+  vsprintf (outstr, fmt, ap);
+  va_end (ap);
+  g_print ("mutt_message: %s\n", outstr);
+#endif
 }
 
-void 
+void
 mutt_exit (int code)
 {
 }
 
-int 
+int
 mutt_yesorno (const char *msg, int def)
 {
   return 1;
 }
 
-int 
+int
 mutt_any_key_to_continue (const char *s)
 {
   return 1;
 }
 
-void 
+void
 mutt_clear_error (void)
 {
 }
+
