@@ -1,5 +1,5 @@
 /* Balsa E-Mail Client
- * Copyright (C) 1997-98 Jay Painter and Stuart Parmenter
+ * Copyright (C) 1997-1999 Stuart Parmenter and Jay Painter
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,75 +20,39 @@
 #include "config.h"
 #include <glib.h>
 #include <stdarg.h>
+/* this should be removed.  it is only used for _() for internationalzation */
+#include <gnome.h>
 
 #include <stdio.h>
 #include <sys/utsname.h>
 #include <sys/stat.h>
 #include <string.h>
 #include <time.h>
-#include <gnome.h>
+
+#ifdef BALSA_USE_THREADS
+#include <pthread.h>
+#endif
 
 #include "mailbackend.h"
 
-#include "balsa-app.h"
-#include "mailbox.h"
+#include "libbalsa.h"
+#include "libbalsa_private.h"
+
 #include "misc.h"
 
+#ifdef BALSA_USE_THREADS
+#include "threads.h"
+#else
+#include "src/save-restore.h" /*config_mailbox_update*/
+#endif
+
 #define BUFFER_SIZE 1024
 
-
-#define LOCK_MAILBOX(mailbox)\
-do {\
-  if (client_mailbox)\
-    {\
-      g_print (_("*** ERROR: Mailbox Lock Failed: %s ***\n"), __PRETTY_FUNCTION__);\
-      return;\
-    }\
-  else\
-    client_mailbox = (mailbox);\
-} while (0)
-
-
-#define LOCK_MAILBOX_RETURN_VAL(mailbox, val)\
-do {\
-  if (client_mailbox)\
-    {\
-      g_print (_("*** ERROR: Mailbox Lock Failed: %s ***\n"), __PRETTY_FUNCTION__);\
-      return (val);\
-    }\
-  else\
-    client_mailbox = (mailbox);\
-} while (0)
-
-#define UNLOCK_MAILBOX()                (client_mailbox = NULL)
-
-
-#define CLIENT_CONTEXT(mailbox)          (((MailboxPrivate *)((mailbox)->private))->context)
-#define CLIENT_CONTEXT_OPEN(mailbox)     (CLIENT_CONTEXT (mailbox) != NULL)
-#define CLIENT_CONTEXT_CLOSED(mailbox)   (CLIENT_CONTEXT (mailbox) == NULL)
-#define RETURN_IF_CLIENT_CONTEXT_CLOSED(mailbox)\
-do {\
-  if (CLIENT_CONTEXT_CLOSED (mailbox))\
-    {\
-      g_print (_("*** ERROR: Mailbox Stream Closed: %s ***\n"), __PRETTY_FUNCTION__);\
-      UNLOCK_MAILBOX ();\
-      return;\
-    }\
-} while (0)
-#define RETURN_VAL_IF_CONTEXT_CLOSED(mailbox, val)\
-do {\
-  if (CLIENT_CONTEXT_CLOSED (mailbox))\
-    {\
-      g_print (_("*** ERROR: Mailbox Stream Closed: %s ***\n"), __PRETTY_FUNCTION__);\
-      UNLOCK_MAILBOX ();\
-      return (val);\
-    }\
-} while (0)
-
-
 #define WATCHER_LIST(mailbox) (((MailboxPrivate *)((mailbox)->private))->watcher_list)
 
-
+/* update the gui (during loading messages, etc */
+void (*update_gui_func) (void);
+	
 /*
  * watcher information
  */
@@ -101,51 +65,51 @@ typedef struct
   }
 MailboxWatcher;
 
-
-/* 
- * private mailbox data
- */
-typedef struct
-  {
-    CONTEXT *context;
-    GList *watcher_list;
-  }
-MailboxPrivate;
-
-
-/* 
- * the mailbox to be referenced by any of the c-client
- * callbacks for authorization, new messages, etc... 
- */
-static Mailbox *client_mailbox = NULL;
+static void balsa_mailbox_class_init (MailboxClass *klass);
+static void balsa_mailbox_init(Mailbox *mailbox);
+static void balsa_mailbox_destroy (GtkObject *object);
+static void balsa_mailbox_real_open_mailbox(Mailbox *mailbox);
+static void balsa_mailbox_real_close_mailbox(Mailbox *mailbox);
 
 
 /* 
  * prototypes
  */
-static void load_messages (Mailbox * mailbox, gint emit);
+/* static void load_messages (Mailbox * mailbox, gint emit); */
 static void free_messages (Mailbox * mailbox);
 
-static void send_watcher_mark_clear_message (Mailbox * mailbox, Message * message);
-static void send_watcher_mark_answer_message (Mailbox * mailbox, Message * message);
-static void send_watcher_mark_read_message (Mailbox * mailbox, Message * message);
-static void send_watcher_mark_unread_message (Mailbox * mailbox, Message * message);
-static void send_watcher_mark_delete_message (Mailbox * mailbox, Message * message);
-static void send_watcher_mark_undelete_message (Mailbox * mailbox, Message * message);
-static void send_watcher_new_message (Mailbox * mailbox, Message * message, gint remaining);
-static void send_watcher_delete_message (Mailbox * mailbox, Message * message);
+
+static gint _mailbox_open_ref (Mailbox * mailbox, gint flag);
+
+/* PKGW: These were static. Why? */
+void send_watcher_mark_clear_message (Mailbox * mailbox, Message * message);
+void send_watcher_mark_answer_message (Mailbox * mailbox, Message * message);
+void send_watcher_mark_read_message (Mailbox * mailbox, Message * message);
+void send_watcher_mark_unread_message (Mailbox * mailbox, Message * message);
+void send_watcher_mark_delete_message (Mailbox * mailbox, Message * message);
+void send_watcher_mark_undelete_message (Mailbox * mailbox, Message * message);
+void send_watcher_new_message (Mailbox * mailbox, Message * message, gint remaining);
+void send_watcher_delete_message (Mailbox * mailbox, Message * message);
+void send_watcher_append_message (Mailbox * mailbox, Message * message);
 
 static Message *translate_message (HEADER * cur);
 static Address *translate_address (ADDRESS * caddr);
 
+Server *server_new(ServerType type);
+void server_free(Server *server);
+
 /* We're gonna set Mutt global vars here */
 void
-mailbox_init (gchar * inbox_path, void (*error_func)(const char* fmt, ...))
+mailbox_init (gchar * inbox_path,
+		void (*error_func) (const char *fmt,...),
+		void (*gui_func) (void))
 {
   struct utsname utsname;
   char *p;
   gchar *tmp;
 
+  update_gui_func = gui_func;
+  
   Spoolfile = inbox_path;
 
   uname (&utsname);
@@ -174,163 +138,256 @@ mailbox_init (gchar * inbox_path, void (*error_func)(const char* fmt, ...))
   snprintf (tmp, 17 + strlen (VERSION), "X-Mailer: Balsa %s", VERSION);
   UserHeader->data = g_strdup (tmp);
   g_free (tmp);
-
 }
 
 gint
 set_imap_username (Mailbox * mb)
 {
-  if (mb->type != MAILBOX_IMAP)
+  if (!MAILBOX_IS_IMAP(mb))
     return 0;
 
-  ImapUser = MAILBOX_IMAP (mb)->user;
-  ImapPass = MAILBOX_IMAP (mb)->passwd;
+  ImapUser = MAILBOX_IMAP(mb)->server->user;
+  ImapPass = MAILBOX_IMAP(mb)->server->passwd;
 
   return 1;
 }
 
+
 void
-check_all_pop3_hosts (Mailbox * to)
+check_all_imap_hosts (Mailbox * to, GList *mailboxes)
+{
+#ifdef BALSA_USE_THREADS
+/*  Only check if lock has been set */
+  pthread_mutex_lock( &mailbox_lock);
+  if( !checking_mail )
+  {
+    pthread_mutex_unlock( &mailbox_lock);
+    return;
+  }
+  pthread_mutex_unlock( &mailbox_lock );
+
+/*  put IMAP code here */
+
+
+  return;  
+#else
+/* non-thread code */
+
+#endif
+}
+
+
+void
+check_all_pop3_hosts (Mailbox *to, GList *mailboxes)
 {
   GList *list;
   Mailbox *mailbox;
+  char uid[80];
 
-  list = g_list_first (balsa_app.inbox_input);
+#ifdef BALSA_USE_THREADS
+  char msgbuf[160];
+  MailThreadMessage *threadmsg;
 
-  if (to->type != MAILBOX_MBOX)
+/*  Only check if lock has been set */
+  pthread_mutex_lock( &mailbox_lock);
+  if( !checking_mail )
+  {
+    pthread_mutex_unlock( &mailbox_lock);
     return;
+  }
+  pthread_mutex_unlock( &mailbox_lock );
+#endif BALSA_USE_THREADS
 
+  list = g_list_first (mailboxes);
+
+ /* if (to->type != MAILBOX_MBOX)
+    return;
+  */
+  
   Spoolfile = MAILBOX_LOCAL (to)->path;
 
   while (list)
+  {
+    mailbox = list->data;
+    if (MAILBOX_POP3 (mailbox)->check)
     {
-      mailbox = list->data;
-      PopHost = g_strdup (MAILBOX_POP3 (mailbox)->server);
+      PopHost = g_strdup (MAILBOX_POP3(mailbox)->server->host);
       PopPort = 110;
-      PopPass = g_strdup (MAILBOX_POP3 (mailbox)->passwd);
-      PopUser = g_strdup (MAILBOX_POP3 (mailbox)->user);
+      PopPass = g_strdup (MAILBOX_POP3(mailbox)->server->passwd);
+      PopUser = g_strdup (MAILBOX_POP3(mailbox)->server->user);
+
+#ifdef BALSA_USE_THREADS
+      sprintf( msgbuf, "POP3: %s", MAILBOX_POP3(mailbox)->mailbox.name );
+      MSGMAILTHREAD( threadmsg, MSGMAILTHREAD_SOURCE, msgbuf );
+#endif
+ 
+      if( MAILBOX_POP3 (mailbox)->last_popped_uid == NULL)
+        uid[0] = 0;
+      else
+        strcpy( uid, MAILBOX_POP3 (mailbox)->last_popped_uid );
+
+      PopUID = uid;
+
+      /* Delete it if necessary */
+      if (MAILBOX_POP3 (mailbox)->delete_from_server)
+      {
+        set_option(OPTPOPDELETE);
+      }
+      else
+      {
+        unset_option(OPTPOPDELETE);
+      }
+
       mutt_fetchPopMail ();
       g_free (PopHost);
       g_free (PopPass);
       g_free (PopUser);
-      list = list->next;
+
+      if(MAILBOX_POP3(mailbox)->last_popped_uid == NULL ||
+         strcmp(MAILBOX_POP3(mailbox)->last_popped_uid, uid) != 0)
+      {
+        g_free ( MAILBOX_POP3 (mailbox)->last_popped_uid );
+        MAILBOX_POP3 (mailbox)->last_popped_uid = g_strdup ( uid );
+
+#ifdef BALSA_USE_THREADS
+	threadmsg = malloc( sizeof( MailThreadMessage ) );
+	threadmsg->message_type = MSGMAILTHREAD_UPDATECONFIG;
+	threadmsg->mailbox = (void *) mailbox;
+	/*  MAILBOX_POP3(mailbox)->mailbox.name */
+        write( mail_thread_pipes[1], (void *) &threadmsg, sizeof(void *) );
+#else
+	config_mailbox_update( mailbox, MAILBOX_POP3(mailbox)->mailbox.name);
+#endif
+      }
     }
+    list = list->next;
+  }
+  return;
 }
 
-#ifdef BUFFY_SIZE
-/*
- * Incoming is a global var in mutt, needs to be set for
- * mailbox_have_new_messages() to work correctly.
- */
-extern int test_new_folder (const char *);
 void
 add_mailboxes_for_checking (Mailbox * mailbox)
 {
   BUFFY **tmp;
-  struct stat sb;
+  gchar *path;
 
-  if (!mailbox)
+  if (!MAILBOX_IS_LOCAL(mailbox))
     return;
 
-  if (mailbox->type == MAILBOX_IMAP ||
-      mailbox->type == MAILBOX_POP3)
-    return;
+  path = MAILBOX_LOCAL (mailbox)->path;
 
   for (tmp = &Incoming; *tmp; tmp = &((*tmp)->next))
     {
-      if (strcmp (MAILBOX_LOCAL (mailbox)->path, (*tmp)->path) == 0)
-	break;
+      if (strcmp (path, (*tmp)->path) == 0)
+	return;
     }
 
   if (!*tmp)
     {
-      *tmp = (BUFFY *) g_malloc (sizeof (BUFFY));
-      (*tmp)->path = g_strdup (MAILBOX_LOCAL (mailbox)->path);
+      *tmp = (BUFFY *) g_new0 (BUFFY, 1);
+      (*tmp)->path = g_strdup (path);
       (*tmp)->next = NULL;
     }
+
   (*tmp)->new = 0;
   (*tmp)->notified = 1;
   (*tmp)->newly_created = 0;
 
-  /* for buffy_size, it is important that if the folder is new (tested
-   * by reading it), the size is set to 0 so that later when we check we see
-   * that it increased .  without buffy_size we probably don't care.
-   */
-  if (stat ((*tmp)->path, &sb) == 0 && !test_new_folder ((*tmp)->path))
+  return;
+}
+
+gint
+mailbox_have_new_messages (gchar * path)
+{
+  BUFFY *tmp = NULL;
+
+  mutt_buffy_notify ();
+
+  for (tmp = Incoming; tmp; tmp = tmp->next)
     {
-      /* some systems out there don't have an off_t type */
-      (*tmp)->size = (long) sb.st_size;
+      if (strcmp (tmp->path, path) == 0)
+	{
+	  if (tmp->new)
+	    return TRUE;
+	}
     }
-  else
-    (*tmp)->size = 0;
 
-  return;
-
-}
-
-gint
-mailbox_have_new_messages (gchar * path)
-{
-  BUFFY *buffy;
-  buffy = mutt_find_mailbox (path);
-  if (buffy)
-    return (buffy->new);
-  else
-    return FALSE;
-}
-#else
-void
-add_mailboxes_for_checking (Mailbox * mailbox)
-{
-  return;
-}
-
-gint
-mailbox_have_new_messages (gchar * path)
-{
   return FALSE;
 }
-#endif /* BUFFY_SIZE */
 
 
-/*
- * allocate a new mailbox
- */
-Mailbox *
-mailbox_new (MailboxType type)
+
+enum {
+  OPEN_MAILBOX,
+  CLOSE_MAILBOX,
+  LAST_SIGNAL
+};
+
+static GtkObjectClass *parent_class = NULL;
+static guint mailbox_signals[LAST_SIGNAL] = { 0 };
+
+GtkType
+balsa_mailbox_get_type (void)
 {
-  Mailbox *mailbox;
+  static GtkType mailbox_type = 0;
 
-  switch (type)
+  if (!mailbox_type)
     {
-    case MAILBOX_MBOX:
-    case MAILBOX_MH:
-    case MAILBOX_MAILDIR:
-      mailbox = (Mailbox *) g_malloc (sizeof (MailboxLocal));
-      MAILBOX_LOCAL (mailbox)->path = NULL;
-      break;
+      static const GtkTypeInfo mailbox_info =
+      {
+	"Mailbox",
+	sizeof (Mailbox),
+	sizeof (MailboxClass),
+	(GtkClassInitFunc) balsa_mailbox_class_init,
+	(GtkObjectInitFunc) balsa_mailbox_init,
+        /* reserved_1 */ NULL,
+	/* reserved_2 */ NULL,
+	(GtkClassInitFunc) NULL,
+      };
 
-    case MAILBOX_POP3:
-      mailbox = (Mailbox *) g_malloc (sizeof (MailboxPOP3));
-      MAILBOX_POP3 (mailbox)->user = NULL;
-      MAILBOX_POP3 (mailbox)->passwd = NULL;
-      MAILBOX_POP3 (mailbox)->server = NULL;
-      break;
-
-    case MAILBOX_IMAP:
-      mailbox = (Mailbox *) g_malloc (sizeof (MailboxIMAP));
-      MAILBOX_IMAP (mailbox)->user = NULL;
-      MAILBOX_IMAP (mailbox)->passwd = NULL;
-      MAILBOX_IMAP (mailbox)->server = NULL;
-      MAILBOX_IMAP (mailbox)->path = NULL;
-      MAILBOX_IMAP (mailbox)->tmp_file_path = NULL;
-      break;
-
-    default:
-      return NULL;
+      mailbox_type = gtk_type_unique (gtk_object_get_type(), &mailbox_info);
     }
 
-  mailbox->type = type;
+  return mailbox_type;
+}
+
+static void
+balsa_mailbox_class_init (MailboxClass *klass)
+{
+  GtkObjectClass *object_class;
+
+  object_class = (GtkObjectClass*) klass;
+
+  parent_class = gtk_type_class(gtk_object_get_type());
+
+  mailbox_signals[OPEN_MAILBOX] =
+    gtk_signal_new ("open_mailbox",
+                    GTK_RUN_FIRST,
+                    object_class->type,
+                    GTK_SIGNAL_OFFSET (MailboxClass, open_mailbox),
+		    gtk_marshal_NONE__NONE,
+		    GTK_TYPE_NONE, 0);
+
+  mailbox_signals[CLOSE_MAILBOX] =
+    gtk_signal_new ("close_mailbox",
+                    GTK_RUN_LAST,
+                    object_class->type,
+                    GTK_SIGNAL_OFFSET (MailboxClass, close_mailbox),
+		    gtk_marshal_NONE__NONE,
+		    GTK_TYPE_NONE, 0);
+
+  gtk_object_class_add_signals (object_class, mailbox_signals, LAST_SIGNAL);
+
+  object_class->destroy = balsa_mailbox_destroy;
+
+  klass->open_mailbox = balsa_mailbox_real_open_mailbox;
+  klass->close_mailbox = balsa_mailbox_real_close_mailbox;
+}
+
+static void
+balsa_mailbox_init(Mailbox *mailbox)
+{
+  mailbox->lock = FALSE;
   mailbox->name = NULL;
   mailbox->private = (void *) g_malloc (sizeof (MailboxPrivate));
   CLIENT_CONTEXT (mailbox) = NULL;
@@ -339,13 +396,42 @@ mailbox_new (MailboxType type)
   mailbox->messages = 0;
   mailbox->new_messages = 0;
   mailbox->message_list = NULL;
-  return mailbox;
 }
 
-
-void
-mailbox_free (Mailbox * mailbox)
+GtkObject*
+mailbox_new(MailboxType type)
 {
+  Mailbox *mailbox;
+
+  switch (type)
+    {
+    case MAILBOX_MBOX:
+    case MAILBOX_MH:
+    case MAILBOX_MAILDIR:
+      mailbox = gtk_type_new(BALSA_TYPE_MAILBOX_LOCAL);
+      break;
+
+    case MAILBOX_POP3:
+      mailbox = gtk_type_new(BALSA_TYPE_MAILBOX_POP3);
+      break;
+
+    case MAILBOX_IMAP:
+      mailbox = gtk_type_new(BALSA_TYPE_MAILBOX_IMAP);
+      break;
+
+    default:
+      return NULL;
+    }
+
+  mailbox->type = type;
+  
+  return GTK_OBJECT(mailbox);
+}
+
+static void balsa_mailbox_destroy (GtkObject *object)
+{
+  Mailbox *mailbox = BALSA_MAILBOX(object);
+
   if (!mailbox)
     return;
 
@@ -353,40 +439,82 @@ mailbox_free (Mailbox * mailbox)
     while (mailbox->open_ref > 0)
       mailbox_open_unref (mailbox);
 
-  g_free (mailbox->name);
-  g_free (mailbox->private);
+  g_free(mailbox->name);
+  g_free(mailbox->private);
 
-  switch (mailbox->type)
-    {
-    case MAILBOX_MBOX:
-    case MAILBOX_MH:
-    case MAILBOX_MAILDIR:
-      g_free (MAILBOX_LOCAL (mailbox)->path);
-      break;
-
-    case MAILBOX_POP3:
-      g_free (MAILBOX_POP3 (mailbox)->user);
-      g_free (MAILBOX_POP3 (mailbox)->passwd);
-      g_free (MAILBOX_POP3 (mailbox)->server);
-      break;
-
-    case MAILBOX_IMAP:
-      g_free (MAILBOX_IMAP (mailbox)->user);
-      g_free (MAILBOX_IMAP (mailbox)->passwd);
-      g_free (MAILBOX_IMAP (mailbox)->server);
-      g_free (MAILBOX_IMAP (mailbox)->path);
-      break;
-
-    case MAILBOX_UNKNOWN:
-      break;
-    }
-
-  g_free (mailbox);
+  if (GTK_OBJECT_CLASS(parent_class)->destroy)
+    (*GTK_OBJECT_CLASS(parent_class)->destroy)(GTK_OBJECT(object));
 }
 
+Server *
+server_new(ServerType type)
+{
+  Server *server;
+
+  /* we can create the same thing for now */
+  server = (Server *)g_malloc(sizeof(Server));
+  server->user = NULL;
+  server->passwd = NULL;
+  server->host = NULL;
+  server->port = -1;
+
+  return server;
+}
+
+void
+server_free(Server *server)
+{
+  g_free(server->user);
+  g_free(server->passwd);
+  g_free(server->host);
+}
+
+
+
+void balsa_mailbox_open(Mailbox *mailbox)
+{
+  g_return_if_fail(mailbox != NULL);
+  g_return_if_fail(BALSA_IS_MAILBOX(mailbox));
+
+  gtk_signal_emit(GTK_OBJECT(mailbox), mailbox_signals[OPEN_MAILBOX], NULL);
+}
+
+void balsa_mailbox_close(Mailbox *mailbox)
+{
+  g_return_if_fail(mailbox != NULL);
+  g_return_if_fail(BALSA_IS_MAILBOX(mailbox));
+
+  gtk_signal_emit(GTK_OBJECT(mailbox), mailbox_signals[CLOSE_MAILBOX], NULL);
+}
+
+
 
 gint
 mailbox_open_ref (Mailbox * mailbox)
+{
+  return _mailbox_open_ref (mailbox, 0);
+}
+
+gint
+mailbox_open_append (Mailbox * mailbox)
+{
+  return _mailbox_open_ref (mailbox, M_APPEND);
+}
+
+
+
+static void balsa_mailbox_real_open_mailbox(Mailbox *mailbox)
+{
+  _mailbox_open_ref(mailbox, 0);
+}
+
+static void balsa_mailbox_real_close_mailbox(Mailbox *mailbox)
+{
+  mailbox_open_unref(mailbox);
+}
+
+static gint
+_mailbox_open_ref (Mailbox * mailbox, gint flag)
 {
   GString *tmp;
   struct stat st;
@@ -395,18 +523,28 @@ mailbox_open_ref (Mailbox * mailbox)
 
   if (CLIENT_CONTEXT_OPEN (mailbox))
     {
-      /* incriment the reference count */
-      mailbox->open_ref++;
-
-      UNLOCK_MAILBOX ();
-      return TRUE;
+      if (flag == M_APPEND)
+	{
+	  
+	  /* we need the mailbox to be opened fresh i think */
+	  mx_close_mailbox( CLIENT_CONTEXT(mailbox)  );
+	  
+	} 
+      else 
+	{
+	  /* incriment the reference count */
+	  mailbox->open_ref++;
+	  
+	  UNLOCK_MAILBOX (mailbox);
+	  return TRUE;
+	}
     }
 
-  if (mailbox->type != MAILBOX_IMAP && mailbox->type != MAILBOX_POP3)
+  if (MAILBOX_IS_LOCAL(mailbox))
     {
       if (stat (MAILBOX_LOCAL (mailbox)->path, &st) == -1)
 	{
-	  UNLOCK_MAILBOX ();
+	  UNLOCK_MAILBOX (mailbox);
 	  return FALSE;
 	}
     }
@@ -417,21 +555,22 @@ mailbox_open_ref (Mailbox * mailbox)
     case MAILBOX_MBOX:
     case MAILBOX_MH:
     case MAILBOX_MAILDIR:
-      CLIENT_CONTEXT (mailbox) = mx_open_mailbox (MAILBOX_LOCAL (mailbox)->path, 0, NULL);
+      CLIENT_CONTEXT (mailbox) = mx_open_mailbox (MAILBOX_LOCAL (mailbox)->path, flag, NULL);
       break;
 
     case MAILBOX_POP3:
-      CLIENT_CONTEXT (mailbox) = mx_open_mailbox (MAILBOX_POP3 (mailbox)->server, 0, NULL);
+      CLIENT_CONTEXT (mailbox) = mx_open_mailbox (MAILBOX_POP3(mailbox)->server->host, flag, NULL);
       break;
 
     case MAILBOX_IMAP:
       tmp = g_string_new (NULL);
       g_string_append_c (tmp, '{');
-      g_string_append (tmp, MAILBOX_IMAP (mailbox)->server);
+      g_string_append (tmp, MAILBOX_IMAP(mailbox)->server->host);
+      g_string_sprintfa (tmp, ":%i", MAILBOX_IMAP(mailbox)->server->port);
       g_string_append_c (tmp, '}');
       g_string_append (tmp, MAILBOX_IMAP (mailbox)->path);
       set_imap_username (mailbox);
-      CLIENT_CONTEXT (mailbox) = mx_open_mailbox (tmp->str, 0, NULL);
+      CLIENT_CONTEXT (mailbox) = mx_open_mailbox (tmp->str, flag, NULL);
       g_string_free (tmp, TRUE);
       break;
 
@@ -445,18 +584,21 @@ mailbox_open_ref (Mailbox * mailbox)
       mailbox->new_messages = CLIENT_CONTEXT (mailbox)->msgcount;
       load_messages (mailbox, 0);
 
-      /* incriment the reference count */
+      /* increment the reference count */
       mailbox->open_ref++;
 
-      if (balsa_app.debug)
-	g_print (_ ("Mailbox: Opening %s Refcount: %d\n"), mailbox->name, mailbox->open_ref);
+#ifdef DEBUG
+      g_print (_ ("Mailbox: Opening %s Refcount: %d\n"), mailbox->name, mailbox->open_ref);
+#endif
 
-      UNLOCK_MAILBOX ();
+      /* FIXME */
+/* mailbox_sort(mailbox, MAILBOX_SORT_DATE); */
+      UNLOCK_MAILBOX (mailbox);
       return TRUE;
     }
   else
     {
-      UNLOCK_MAILBOX ();
+      UNLOCK_MAILBOX (mailbox);
       return FALSE;
     }
 }
@@ -465,6 +607,9 @@ mailbox_open_ref (Mailbox * mailbox)
 void
 mailbox_open_unref (Mailbox * mailbox)
 {
+#ifdef DEBUG
+      g_print (_ ("Mailbox: Closing %s Refcount: %d\n"), mailbox->name, mailbox->open_ref);
+#endif
   LOCK_MAILBOX (mailbox);
 
   if (mailbox->open_ref == 0)
@@ -474,9 +619,6 @@ mailbox_open_unref (Mailbox * mailbox)
 
   if (mailbox->open_ref == 0)
     {
-      if (balsa_app.debug)
-	g_print (_ ("Mailbox: Closing %s Refcount: %d\n"), mailbox->name, mailbox->open_ref);
-
       free_messages (mailbox);
       mailbox->messages = 0;
 
@@ -488,17 +630,27 @@ mailbox_open_unref (Mailbox * mailbox)
 	     don't ask me what to do - AC */
 
 	  if (mx_close_mailbox (CLIENT_CONTEXT (mailbox)) == 0)
-	    CLIENT_CONTEXT (mailbox) = NULL;
+	    {
+	      free (CLIENT_CONTEXT (mailbox));
+	      CLIENT_CONTEXT (mailbox) = NULL;
+	    }
 	}
     }
 
-  UNLOCK_MAILBOX ();
+  UNLOCK_MAILBOX (mailbox);
+}
+
+void
+mailbox_sort (Mailbox * mailbox, MailboxSort sort)
+{
+  mutt_sort_headers (CLIENT_CONTEXT (mailbox), sort);
 }
 
 gint
 mailbox_check_new_messages (Mailbox * mailbox)
 {
   gint i = 0;
+  gint index_hint;
 
   if (!mailbox)
     return FALSE;
@@ -506,32 +658,44 @@ mailbox_check_new_messages (Mailbox * mailbox)
   LOCK_MAILBOX_RETURN_VAL (mailbox, FALSE);
   RETURN_VAL_IF_CONTEXT_CLOSED (mailbox, FALSE);
 
-  if ((i = mx_check_mailbox (CLIENT_CONTEXT (mailbox), NULL)) < 0)
+  index_hint = CLIENT_CONTEXT (mailbox)->vcount;
+
+  if ((i = mx_check_mailbox (CLIENT_CONTEXT (mailbox), &index_hint)) < 0)
     {
-      UNLOCK_MAILBOX ();
+      UNLOCK_MAILBOX (mailbox);
       g_print ("error or something\n");
     }
   else if (i == M_NEW_MAIL || i == M_REOPENED)
     {
-
+      // g_print ("got new mail! yippie!\n");
       mailbox->new_messages = CLIENT_CONTEXT (mailbox)->msgcount - mailbox->messages;
 
       if (mailbox->new_messages > 0)
 	{
+
+	  mailbox->unread_messages += mailbox->new_messages ;
+	  mailbox->total_messages += mailbox->new_messages ;
+	  /* TODO:the preceeding two lines should be put in load_messages 
+	     but I don't want to rely on the 'emit' flag to know if there is REALLY 
+	     new mail in the mailbox. -bertrand */
+
+#ifndef BALSA_USE_THREADS
 	  load_messages (mailbox, 1);
-	  UNLOCK_MAILBOX ();
+#endif
+	  UNLOCK_MAILBOX(mailbox);
 	  return TRUE;
 	}
       else
 	{
-	  UNLOCK_MAILBOX ();
+	  UNLOCK_MAILBOX (mailbox);
 	  return FALSE;
 	}
     }
-  UNLOCK_MAILBOX ();
+  UNLOCK_MAILBOX (mailbox);
   return FALSE;
 }
 
+
 guint
 mailbox_watcher_set (Mailbox * mailbox,
 		     MailboxWatcherFunc func,
@@ -635,7 +799,7 @@ mailbox_watcher_remove_by_data (Mailbox * mailbox, gpointer data)
 /*
  * private
  */
-static void
+void
 load_messages (Mailbox * mailbox, gint emit)
 {
   glong msgno;
@@ -650,6 +814,13 @@ load_messages (Mailbox * mailbox, gint emit)
 
       if (!cur)
 	continue;
+      if (cur->env->subject && 
+          !strcmp (cur->env->subject, 
+                   "DON'T DELETE THIS MESSAGE -- FOLDER INTERNAL DATA")) {
+          mailbox->new_messages--;
+          mailbox->messages++;
+          continue;
+      }
 
       message = translate_message (cur);
       message->mailbox = mailbox;
@@ -670,16 +841,9 @@ load_messages (Mailbox * mailbox, gint emit)
 
       mailbox->message_list = g_list_append (mailbox->message_list, message);
       mailbox->new_messages--;
-
+     
       if (emit)
-	send_watcher_new_message (mailbox, message, mailbox->new_messages);
-
-      /* 
-       * give time to gtk so the GUI isn't blocked
-       * this is kinda a hack right now
-       */
-      while (gtk_events_pending ())
-	gtk_main_iteration ();
+        send_watcher_new_message (mailbox, message, mailbox->new_messages);
     }
 }
 
@@ -697,7 +861,7 @@ free_messages (Mailbox * mailbox)
       list = list->next;
 
       send_watcher_delete_message (mailbox, message);
-      message_free (message);
+      message_destroy (message);
     }
   g_list_free (mailbox->message_list);
   mailbox->message_list = NULL;
@@ -707,7 +871,7 @@ free_messages (Mailbox * mailbox)
 /*
  * sending messages to watchers
  */
-static void
+void
 send_watcher_mark_clear_message (Mailbox * mailbox, Message * message)
 {
   GList *list;
@@ -732,7 +896,7 @@ send_watcher_mark_clear_message (Mailbox * mailbox, Message * message)
     }
 }
 
-static void
+void
 send_watcher_mark_answer_message (Mailbox * mailbox, Message * message)
 {
   GList *list;
@@ -757,7 +921,7 @@ send_watcher_mark_answer_message (Mailbox * mailbox, Message * message)
     }
 }
 
-static void
+void
 send_watcher_mark_read_message (Mailbox * mailbox, Message * message)
 {
   GList *list;
@@ -782,7 +946,7 @@ send_watcher_mark_read_message (Mailbox * mailbox, Message * message)
     }
 }
 
-static void
+void
 send_watcher_mark_unread_message (Mailbox * mailbox, Message * message)
 {
   GList *list;
@@ -807,7 +971,7 @@ send_watcher_mark_unread_message (Mailbox * mailbox, Message * message)
     }
 }
 
-static void
+void
 send_watcher_mark_delete_message (Mailbox * mailbox, Message * message)
 {
   GList *list;
@@ -833,7 +997,7 @@ send_watcher_mark_delete_message (Mailbox * mailbox, Message * message)
 }
 
 
-static void
+void
 send_watcher_mark_undelete_message (Mailbox * mailbox, Message * message)
 {
   GList *list;
@@ -859,7 +1023,7 @@ send_watcher_mark_undelete_message (Mailbox * mailbox, Message * message)
 }
 
 
-static void
+void
 send_watcher_new_message (Mailbox * mailbox, Message * message, gint remaining)
 {
   GList *list;
@@ -886,8 +1050,34 @@ send_watcher_new_message (Mailbox * mailbox, Message * message, gint remaining)
     }
 }
 
+void
+send_watcher_append_message (Mailbox * mailbox, Message * message)
+{
+  GList *list;
+  MailboxWatcherMessage  mw_message;
+  MailboxWatcher *watcher;
 
-static void
+  mw_message.type = MESSAGE_APPEND;
+  mw_message.mailbox = mailbox;
+  mw_message.message = message;
+
+
+  list = WATCHER_LIST (mailbox);
+  while (list)
+    {
+      watcher = list->data;
+      list = list->next;
+
+      if (watcher->mask & MESSAGE_APPEND_MASK)
+	{
+	  mw_message.data = watcher->data;
+	  (*watcher->func) ((MailboxWatcherMessage *) & mw_message);
+	}
+    }
+}
+
+
+void
 send_watcher_delete_message (Mailbox * mailbox, Message * message)
 {
   GList *list;
@@ -1003,216 +1193,76 @@ mailbox_valid (gchar * filename)
     }
 }
 
-
-/*
- * messages
- */
-Message *
-message_new (void)
-{
-  Message *message;
-
-  message = g_malloc (sizeof (Message));
-
-  message->flags = 0;
-  message->msgno = 0;
-  message->mailbox = NULL;
-  message->remail = NULL;
-  message->date = NULL;
-  message->from = NULL;
-  message->sender = NULL;
-  message->reply_to = NULL;
-  message->subject = NULL;
-  message->to_list = NULL;
-  message->cc_list = NULL;
-  message->bcc_list = NULL;
-  message->in_reply_to = NULL;
-  message->message_id = NULL;
-  message->body_ref = 0;
-  message->body_list = NULL;
-
-  return message;
-}
-
-
-void
-message_free (Message * message)
-{
-  g_free (message->remail);
-  g_free (message->date);
-  address_free (message->from);
-  address_free (message->sender);
-  address_free (message->reply_to);
-  g_free (message->subject);
-
-  g_list_free (message->to_list);
-  g_list_free (message->cc_list);
-
-  g_free (message->in_reply_to);
-  g_free (message->message_id);
-
-  /* finally free the message */
-  g_free (message);
-}
-
-gchar *
-message_pathname (Message * message)
-{
-  return CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno]->path;
-}
-
-/* TODO and/or FIXME look at the flags for mutt_append_message */
-void
-message_copy (Message * message, Mailbox * dest)
-{
-  HEADER *cur;
-
-  LOCK_MAILBOX (message->mailbox);
-  RETURN_IF_CLIENT_CONTEXT_CLOSED (message->mailbox);
-
-  cur = CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno];
-
-  mailbox_open_ref (dest);
-
-  mutt_append_message (CLIENT_CONTEXT (dest),
-		       CLIENT_CONTEXT (message->mailbox),
-		       cur, 0, 0);
-
-  mailbox_open_unref (dest);
-
-  UNLOCK_MAILBOX ();
-}
-
-void
-message_move (Message * message, Mailbox * dest)
-{
-  HEADER *cur;
-
-  LOCK_MAILBOX (message->mailbox);
-  RETURN_IF_CLIENT_CONTEXT_CLOSED (message->mailbox);
-
-  cur = CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno];
-
-  mailbox_open_ref (dest);
-
-  mutt_append_message (CLIENT_CONTEXT (dest),
-		       CLIENT_CONTEXT (message->mailbox),
-		       cur, 0, 0);
-
-  mailbox_open_unref (dest);
-
-  UNLOCK_MAILBOX ();
-
-  message_delete (message);
-}
-
-void
-message_reply (Message * message)
-{
-  HEADER *cur;
-
-  LOCK_MAILBOX (message->mailbox);
-  RETURN_IF_CLIENT_CONTEXT_CLOSED (message->mailbox);
-
-  cur = CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno];
-
-  mutt_set_flag (CLIENT_CONTEXT (message->mailbox), cur, M_REPLIED, TRUE);
-
-  message->flags |= MESSAGE_FLAG_REPLIED;
-  send_watcher_mark_answer_message (message->mailbox, message);
-
-  UNLOCK_MAILBOX ();
-}
-
-void
-message_clear_flags (Message * message)
+/**
+ * mailbox_gather_content_info:
+ *
+ * @mailbox : the mailbox to scan
+ *
+ * gather informations about the content of a mailbox such as the
+ * total nuber of messages, the number of unread messages 
+ * 
+ *
+ * Return value: if the mailbox could be scanned, returns true. 
+ **/
+gboolean
+mailbox_gather_content_info(Mailbox *mailbox)
 {
+	/* this code is far too slow, and mut does not provide a good way to
+	 * do this.  we will not use it for now */
 #if 0
-  char tmp[BUFFER_SIZE];
+  GList *message_list;
+  Message *current_message;
 
-  LOCK_MAILBOX (message->mailbox);
-  RETURN_IF_CLIENT_CONTEXT_CLOSED (message->mailbox);
+  mailbox_open_ref (mailbox);
 
-  mutt_set_flag (CLIENT_CONTEXT (message->mailbox), cur, M_REPLIED, 1);
-  sprintf (tmp, "%ld", message->msgno);
-  mail_clearflag (CLIENT_STREAM (message->mailbox), tmp, "\\DELETED");
-  mail_clearflag (CLIENT_STREAM (message->mailbox), tmp, "\\ANSWERED");
+  mailbox->total_messages = 0;
+  mailbox->unread_messages = 0;
+  /* examine all the message in the mailbox */
+  message_list = mailbox->message_list;
+  while (message_list)
+    {
+      current_message = (Message *) message_list->data;
+      if ( current_message->flags & MESSAGE_FLAG_NEW ) mailbox->unread_messages++ ;
+      mailbox->total_messages++ ;
+      message_list = message_list->next;
+      
+    }
 
-  message->flags = 0;
-  send_watcher_mark_clear_message (message->mailbox, message);
-
-  UNLOCK_MAILBOX ();
+  mailbox_open_unref (mailbox);
 #endif
+  return TRUE;
 }
 
 
-void
-message_read (Message * message)
+void mailbox_commit_flagged_changes( Mailbox *mailbox )
 {
-  HEADER *cur = CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno];
+  GList *message_list;
+  GList *tmp_message_list;
+  Message *current_message;
 
-  LOCK_MAILBOX (message->mailbox);
-  RETURN_IF_CLIENT_CONTEXT_CLOSED (message->mailbox);
 
-  mutt_set_flag (CLIENT_CONTEXT (message->mailbox), cur, M_READ, TRUE);
-  mutt_set_flag (CLIENT_CONTEXT (message->mailbox), cur, M_OLD, FALSE);
+  mailbox_open_ref (mailbox);
 
-  message->flags &= ~MESSAGE_FLAG_NEW;
-  send_watcher_mark_read_message (message->mailbox, message);
+  /* examine all the message in the mailbox */
+  message_list = mailbox->message_list;
+  while (message_list)
+    {
+      current_message = (Message *) message_list->data;
+      tmp_message_list =  message_list->next;
+      if ( current_message->flags & MESSAGE_FLAG_DELETED ) 
+	{
+	   send_watcher_delete_message (mailbox, current_message);
+	   message_destroy (current_message);
+	   mailbox->message_list = g_list_remove_link( mailbox->message_list, message_list);
+	}
+      message_list = tmp_message_list;
+      
+    }
 
-  UNLOCK_MAILBOX ();
+  mailbox_open_unref (mailbox);
+  //  mx_sync_mailbox( CLIENT_CONTEXT(mailbox)  );
 }
 
-void
-message_unread (Message * message)
-{
-  HEADER *cur = CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno];
-
-  LOCK_MAILBOX (message->mailbox);
-  RETURN_IF_CLIENT_CONTEXT_CLOSED (message->mailbox);
-
-  mutt_set_flag (CLIENT_CONTEXT (message->mailbox), cur, M_READ, TRUE);
-
-  message->flags |= MESSAGE_FLAG_NEW;
-  send_watcher_mark_unread_message (message->mailbox, message);
-
-  UNLOCK_MAILBOX ();
-}
-
-void
-message_delete (Message * message)
-{
-  HEADER *cur = CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno];
-
-  LOCK_MAILBOX (message->mailbox);
-  RETURN_IF_CLIENT_CONTEXT_CLOSED (message->mailbox);
-
-  mutt_set_flag (CLIENT_CONTEXT (message->mailbox), cur, M_DELETE, TRUE);
-
-  message->flags |= MESSAGE_FLAG_DELETED;
-  send_watcher_mark_delete_message (message->mailbox, message);
-
-  UNLOCK_MAILBOX ();
-}
-
-
-void
-message_undelete (Message * message)
-{
-  HEADER *cur = CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno];
-
-  LOCK_MAILBOX (message->mailbox);
-  RETURN_IF_CLIENT_CONTEXT_CLOSED (message->mailbox);
-
-  mutt_set_flag (CLIENT_CONTEXT (message->mailbox), cur, M_DELETE, FALSE);
-
-  message->flags &= ~MESSAGE_FLAG_DELETED;
-  send_watcher_mark_undelete_message (message->mailbox, message);
-
-  UNLOCK_MAILBOX ();
-}
-
-
 /* internal c-client translation */
 static Message *
 translate_message (HEADER * cur)
@@ -1230,13 +1280,12 @@ translate_message (HEADER * cur)
   cenv = cur->env;
 
   message = message_new ();
-/*
-   message->remail = g_strdup (cenv->remail);
- */
+
   footime = localtime (&cur->date_sent);
 
   strftime (rettime, sizeof (rettime), "%a, %d %b %Y %H:%M:%S", footime);
 
+  message->datet = cur->date_sent;
   message->date = g_strdup (rettime);
   message->from = translate_address (cenv->from);
   message->sender = translate_address (cenv->sender);
@@ -1269,164 +1318,6 @@ translate_message (HEADER * cur)
 
 
 
-char *
-mime_content_type2str (int contenttype)
-{
-  switch (contenttype)
-    {
-    case TYPEOTHER:
-      return "OTHER";
-    case TYPEAUDIO:
-      return "AUDIO";
-    case TYPEAPPLICATION:
-      return "APPLICATION";
-    case TYPEIMAGE:
-      return "IMAGE";
-    case TYPEMULTIPART:
-      return "MULTIPART";
-    case TYPETEXT:
-      return "TEXT";
-    case TYPEVIDEO:
-      return "VIDEO";
-    default:
-      return "";
-    }
-}
-
-
-
-
-
-void
-message_body_ref (Message * message)
-{
-  Body *body;
-  HEADER *cur;
-  MESSAGE *msg;
-
-  if (!message)
-    return;
-
-  cur = CLIENT_CONTEXT (message->mailbox)->hdrs[message->msgno];
-
-  if (message->body_ref > 0)
-    {
-      message->body_ref++;
-      return;
-    }
-
-  /*
-   * load message body
-   */
-  msg = mx_open_message (CLIENT_CONTEXT (message->mailbox), cur->msgno);
-  if (!msg)
-    {
-      message->body_ref--;
-      return;
-    }
-  fseek (msg->fp, cur->content->offset, 0);
-  if (cur->content->type == TYPEMULTIPART)
-    {
-      cur->content->parts = mutt_parse_multipart (msg->fp,
-		   mutt_get_parameter ("boundary", cur->content->parameter),
-				cur->content->offset + cur->content->length,
-			 strcasecmp ("digest", cur->content->subtype) == 0);
-    }
-  if (msg != NULL)
-    {
-      BODY *bdy = cur->content;
-
-      if (balsa_app.debug)
-	g_print (_ ("Loading message: %s/%s\n"), TYPE (bdy->type), bdy->subtype);
-
-      if (balsa_app.debug)
-	{
-	  fprintf (stderr, "After loading message\n");
-	  fprintf (stderr, "header->mime    = %d\n", cur->mime);
-	  fprintf (stderr, "header->offset  = %ld\n", cur->offset);
-	  fprintf (stderr, "header->content = %p\n", cur->content);
-	  fprintf (stderr, " \n\nDumping Message\n");
-	  fprintf (stderr, "Dumping a %s[%d] message\n",
-		   mime_content_type2str (cur->content->type),
-		   cur->content->type);
-	}
-      if (cur->content->type != TYPEMULTIPART)
-	{
-	  body = body_new ();
-	  body->mutt_body = cur->content;
-	  message->body_list = g_list_append (message->body_list, body);
-	}
-      else
-	{
-	  bdy = cur->content->parts;
-	  while (bdy)
-	    {
-
-	      if (balsa_app.debug)
-		{
-		  fprintf (stderr, "h->c->type      = %s[%d]\n", mime_content_type2str (bdy->type), bdy->type);
-		  fprintf (stderr, "h->c->subtype   = %s\n", bdy->subtype);
-		  fprintf (stderr, "======\n");
-		}
-	      body = body_new ();
-	      body->mutt_body = bdy;
-	      message->body_list = g_list_append (message->body_list, body);
-	      bdy = bdy->next;
-
-	    }
-	}
-      message->body_ref++;
-      mx_close_message (&msg);
-    }
-  /*
-   * emit read message
-   */
-  if (message->flags & MESSAGE_FLAG_NEW)
-    {
-      message_read (message);
-    }
-
-  if (message->mailbox->type == MAILBOX_IMAP)
-    {
-      if (MAILBOX_IMAP (message->mailbox)->tmp_file_path)
-	g_free (MAILBOX_IMAP (message->mailbox)->tmp_file_path);
-      MAILBOX_IMAP (message->mailbox)->tmp_file_path =
-	g_strdup (cur->content->filename);
-    }
-}
-
-
-void
-message_body_unref (Message * message)
-{
-  GList *list;
-  Body *body;
-
-  if (!message)
-    return;
-
-  if (message->body_ref == 0)
-    return;
-
-  if (--message->body_ref == 0)
-    {
-      list = message->body_list;
-      while (list)
-	{
-	  body = (Body *) list->data;
-	  list = list->next;
-
-	  if (body->htmlized)
-	    g_free (body->htmlized);
-	  if (body->buffer)
-	    g_free (body->buffer);
-	}
-
-      g_list_free (message->body_list);
-      message->body_list = NULL;
-    }
-}
-
 /*
  * addresses
  */
@@ -1505,6 +1396,7 @@ body_new ()
   body->htmlized = NULL;
   body->buffer = NULL;
   body->mutt_body = NULL;
+  body->filename = NULL;
   return body;
 }
 
@@ -1515,9 +1407,8 @@ body_free (Body * body)
   if (!body)
     return;
 
-  if (body->htmlized)
-    g_free (body->htmlized);
-  if (body->buffer)
-    g_free (body->buffer);
+  g_free (body->htmlized);
+  g_free (body->buffer);
+  g_free (body->filename);
   g_free (body);
 }
