@@ -1,5 +1,5 @@
 /* Balsa E-Mail Client
- * Copyright (C) 1997-98 Jay Painter and Stuart Parmenter
+ * Copyright (C) 1997-1999 Jay Painter and Stuart Parmenter
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,28 +20,23 @@
 #ifndef __MISC_H__
 #define __MISC_H__
 
+#include "libbalsa.h"
 
-GtkWidget *append_menuitem_connect (GtkMenu * menu,
-				    gchar * text,
-				    GtkSignalFunc func,
-				    gpointer data,
-				    gpointer user_data);
-
-gint g_list_index (GList * list, gpointer data);
-
-gchar *address_to_gchar(Address *addr);
+gchar *address_to_gchar (Address * addr);
 gchar *make_string_from_list (GList *);
 
 size_t readfile (FILE * fp, char **buf);
 
 typedef struct _MailboxNode MailboxNode;
 struct _MailboxNode
-  {
-    gchar *name;
-    Mailbox *mailbox;
-    gint IsDir;
-  };
+{
+  GtkObject object;
+  gchar *name;
+  Mailbox *mailbox;
+  gint IsDir;
+  gint expanded;
+};
 
-MailboxNode *mailbox_node_new (gchar * name, Mailbox * mb, gint i);
-gchar *g_get_host_name(void);
+MailboxNode *mailbox_node_new (const gchar * name, Mailbox * mb, gint i);
+gchar *g_get_host_name (void);
 #endif /* __MISC_H__ */
