@@ -1,5 +1,5 @@
 /* Balsa E-Mail Client
- * Copyright (C) 1997-98 Jay Painter and Stuart Parmenter
+ * Copyright (C) 1997-1999 Jay Painter and Stuart Parmenter
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,8 @@ extern "C"
 {
 #endif				/* __cplusplus */
 
-gboolean send_message (Message * message, gchar * smtp_server, glong debug);
+  gboolean balsa_send_message (Message * message);
+  gboolean balsa_postpone_message (Message * message);
 
 
 #ifdef __cplusplus
