@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 1996-8 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1999 Thomas Roessler <roessler@guug.de>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -23,29 +24,80 @@
 #include "rfc1524.h"
 #include "mime.h"
 #include "pager.h"
+#include "mailbox.h"
+#include "copy.h"
+#include "mx.h"
+
+#ifdef _PGPPATH
+#include "pgp.h"
+#endif
 
 #include <ctype.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <string.h>
 #include <errno.h>
 
 #ifndef LIBMUTT
+int mutt_get_tmp_attachment (BODY *a)
+{
+  char type[STRING];
+  char tempfile[_POSIX_PATH_MAX];
+  rfc1524_entry *entry = rfc1524_new_entry();
+  FILE *fpin = NULL, *fpout = NULL;
+  struct stat st;
+  
+  if(a->unlink)
+    return 0;
+
+  snprintf(type, sizeof(type), "%s/%s", TYPE(a), a->subtype);
+  rfc1524_mailcap_lookup(a, type, entry, 0);
+  rfc1524_expand_filename(entry->nametemplate, a->filename, 
+			  tempfile, sizeof(tempfile));
+  
+  rfc1524_free_entry(&entry);
+
+  if(stat(a->filename, &st) == -1)
+    return -1;
+  
+  if((fpin = fopen(a->filename, "r")) && (fpout = safe_fopen(tempfile, "w")))
+  {
+    mutt_copy_stream (fpin, fpout);
+    FREE(&a->filename);
+    a->filename = safe_strdup(tempfile);
+    a->unlink = 1;
+
+    if(a->stamp >= st.st_mtime)
+      mutt_stamp_attachment(a);
+  }
+  else
+    mutt_perror(fpin ? tempfile : a->filename);
+  
+  if(fpin)  fclose(fpin);
+  if(fpout) fclose(fpout);
+  
+  return a->unlink ? 0 : -1;
+}
+
+
 /* return 1 if require full screen redraw, 0 otherwise */
 int mutt_compose_attachment (BODY *a)
 {
   char type[STRING];
   char command[STRING];
+  char newfile[_POSIX_PATH_MAX] = "";
   rfc1524_entry *entry = rfc1524_new_entry ();
-
-  snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);
+  short unlink_newfile = 0;
+  int rc = 0;
+  
+  snprintf (type, sizeof (type), "%s/%s", TYPE (a), a->subtype);
   if (rfc1524_mailcap_lookup (a, type, entry, M_COMPOSE))
   {
     if (entry->composecommand || entry->composetypecommand)
     {
-      char newfile[_POSIX_PATH_MAX] = "";
 
       if (entry->composetypecommand)
 	strfcpy (command, entry->composetypecommand, sizeof (command));
@@ -56,19 +108,17 @@ int mutt_compose_attachment (BODY *a)
       {
 	dprint(1, (debugfile, "oldfile: %s\t newfile: %s\n",
 				  a->filename, newfile));
-	if (!mutt_rename_file (a->filename, newfile))
+	if (safe_symlink (a->filename, newfile) == -1)
 	{
 	  if (!mutt_yesorno ("Can't match nametemplate, continue?", 1))
-	    return 0;
-	}
-	else
-	{
-	  safe_free ((void **) &a->filename);
-	  a->filename = safe_strdup (newfile);
+	    goto bailout;
 	}
+	unlink_newfile = 1;
       }
-
-      if (rfc1524_expand_command (a, a->filename, type,
+      else
+	strfcpy(newfile, a->filename, sizeof(newfile));
+      
+      if (rfc1524_expand_command (a, newfile, type,
 				      command, sizeof (command)))
       {
 	/* For now, editing requires a file, no piping */
@@ -76,9 +126,13 @@ int mutt_compose_attachment (BODY *a)
       }
       else
       {
+	int r;
+
 	endwin ();
-	mutt_system (command);
-	if (entry->composetypecommand)
+	if ((r = mutt_system (command)) == -1)
+	  mutt_error ("Error running \"%s\"!", command);
+	
+	if (r != -1 && entry->composetypecommand)
 	{
 	  BODY *b;
 	  FILE *fp, *tfp;
@@ -87,7 +141,7 @@ int mutt_compose_attachment (BODY *a)
 	  if ((fp = safe_fopen (a->filename, "r")) == NULL)
 	  {
 	    mutt_perror ("Failure to open file to parse headers.");
-	    return 0;
+	    goto bailout;
 	  }
 
 	  b = mutt_read_mime_header (fp, 0);
@@ -118,7 +172,7 @@ int mutt_compose_attachment (BODY *a)
 	    if ((tfp = safe_fopen (tempfile, "w")) == NULL)
 	    {
 	      mutt_perror ("Failure to open file to strip headers.");
-	      return 0;
+	      goto bailout;
 	    }
 	    mutt_copy_stream (fp, tfp);
 	    fclose (fp);
@@ -135,12 +189,20 @@ int mutt_compose_attachment (BODY *a)
   else
   {
     rfc1524_free_entry (&entry);
-    mutt_message ("No mailcap compose entry for %s, creating empty file.",type);
+    mutt_message ("No mailcap compose entry for %s, creating empty file.",
+		   type);
     return 1;
   }
 
+  rc = 1;
+  
+  bailout:
+  
+  if(unlink_newfile)
+    unlink(newfile);
+
   rfc1524_free_entry (&entry);
-  return 1;
+  return rc;
 }
 
 /* 
@@ -152,18 +214,20 @@ int mutt_compose_attachment (BODY *a)
  * Returns 1 if editor found, 0 if not (useful to tell calling menu to
  * redraw)
  */
-int mutt_edit_attachment (BODY *a, int opt)
+int mutt_edit_attachment (BODY *a)
 {
   char type[STRING];
   char command[STRING];
+  char newfile[_POSIX_PATH_MAX] = "";
   rfc1524_entry *entry = rfc1524_new_entry ();
-
-  snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);
+  short unlink_newfile = 0;
+  int rc = 0;
+  
+  snprintf (type, sizeof (type), "%s/%s", TYPE (a), a->subtype);
   if (rfc1524_mailcap_lookup (a, type, entry, M_EDIT))
   {
     if (entry->editcommand)
     {
-      char newfile[_POSIX_PATH_MAX] = "";
 
       strfcpy (command, entry->editcommand, sizeof (command));
       if (rfc1524_expand_filename (entry->nametemplate,
@@ -171,19 +235,17 @@ int mutt_edit_attachment (BODY *a, int opt)
       {
 	dprint(1, (debugfile, "oldfile: %s\t newfile: %s\n",
 				  a->filename, newfile));
-	if (mutt_rename_file (a->filename, newfile))
+	if (safe_symlink (a->filename, newfile) == -1)
 	{
 	  if (!mutt_yesorno ("Can't match nametemplate, continue?", 1))
-	    return 0;
-	}
-	else
-	{
-	  safe_free ((void **) &a->filename);
-	  a->filename = safe_strdup (newfile);
+	    goto bailout;
 	}
+	unlink_newfile = 1;
       }
+      else
+	strfcpy(newfile, a->filename, sizeof(newfile));
 
-      if (rfc1524_expand_command (a, a->filename, type,
+      if (rfc1524_expand_command (a, newfile, type,
 				      command, sizeof (command)))
       {
 	/* For now, editing requires a file, no piping */
@@ -192,15 +254,16 @@ int mutt_edit_attachment (BODY *a, int opt)
       else
       {
 	endwin ();
-	mutt_system (command);
+	if (mutt_system (command) == -1)
+	  mutt_error ("Error running \"%s\"!", command);
       }
     }
   }
   else if (a->type == TYPETEXT)
   {
     /* On text, default to editor */
-    mutt_edit_file (strcmp ("builtin", Editor) == 0 ? Visual : Editor, 
-		    a->filename);
+    mutt_edit_file ((!Editor || mutt_strcmp ("builtin", Editor) == 0) ? 
+		    NONULL(Visual) : NONULL(Editor), newfile);
   }
   else
   {
@@ -209,28 +272,96 @@ int mutt_edit_attachment (BODY *a, int opt)
     return 0;
   }
 
+  rc = 1;
+  
+  bailout:
+  
+  if(unlink_newfile)
+    unlink(newfile);
+  
   rfc1524_free_entry (&entry);
-  return 1;
+  return rc;
 }
 #endif
-int mutt_is_autoview (char *type)
+
+
+/* for compatibility with metamail */
+static int is_mmnoask (const char *buf)
+{
+  char tmp[LONG_STRING], *p, *q;
+  int lng;
+
+  if ((p = getenv ("MM_NOASK")) != NULL && *p)
+  {
+    if (mutt_strcmp (p, "1") == 0)
+      return (1);
+
+    strfcpy (tmp, p, sizeof (tmp));
+    p = tmp;
+
+    while ((p = strtok (p, ",")) != NULL)
+    {
+      if ((q = strrchr (p, '/')) != NULL)
+      {
+	if (*(q+1) == '*')
+	{
+	  if (mutt_strncasecmp (buf, p, q-p) == 0)
+	    return (1);
+	}
+	else
+	{
+	  if (mutt_strcasecmp (buf, p) == 0)
+	    return (1);
+	}
+      }
+      else
+      {
+	lng = mutt_strlen (p);
+	if (buf[lng] == '/' && mutt_strncasecmp (buf, p, lng) == 0)
+	  return (1);
+      }
+
+      p = NULL;
+    }
+  }
+
+  return (0);
+}
+
+int mutt_is_autoview (BODY *b, const char *type)
 {
   LIST *t = AutoViewList;
+  char _type[SHORT_STRING];
   int i;
 
-  while (t)
+  if (!type)
   {
-    i = strlen (t->data) - 1;
+    snprintf (_type, sizeof (_type), "%s/%s", TYPE (b), b->subtype);
+    type = _type;
+  }
+
+  if (mutt_needs_mailcap (b))
+  {
+    if (option (OPTIMPLICITAUTOVIEW))
+      return 1;
+    
+    if (is_mmnoask (type))
+      return 1;
+  }
+
+  for (; t; t = t->next)
+  {
+    i = mutt_strlen (t->data) - 1;
     if ((i > 0 && t->data[i-1] == '/' && t->data[i] == '*' && 
-	  strncasecmp (type, t->data, i) == 0) ||
-	  strcasecmp (type, t->data) == 0)
+	  mutt_strncasecmp (type, t->data, i) == 0) ||
+	  mutt_strcasecmp (type, t->data) == 0)
       return 1;
-    t = t->next;
   }
 
   return 0;
 }
 #ifndef LIBMUTT
+
 /* returns -1 on error, 0 or the return code from mutt_do_pager() on success */
 int mutt_view_attachment (FILE *fp, BODY *a, int flag)
 {
@@ -243,15 +374,19 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
   char type[STRING];
   char command[STRING];
   char descrip[STRING];
+  char *fname;
   rfc1524_entry *entry = NULL;
   int rc = -1;
-
-  is_message = (a->type == TYPEMESSAGE && a->subtype &&
-		  (!strcasecmp (a->subtype,"rfc822") ||
-		   !strcasecmp (a->subtype, "news")));
+  int unlink_tempfile = 0;
+  
+  is_message = mutt_is_message_type(a->type, a->subtype);
+#ifdef _PGPPATH
+  if (is_message && a->hdr && (a->hdr->pgp & PGPENCRYPT) && !pgp_valid_passphrase())
+    return (rc);
+#endif /* _PGPPATH */
   use_mailcap = (flag == M_MAILCAP ||
 		(flag == M_REGULAR && mutt_needs_mailcap (a)));
-  snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);
+  snprintf (type, sizeof (type), "%s/%s", TYPE (a), a->subtype);
   
   if (use_mailcap)
   {
@@ -280,13 +415,21 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
     }
     strfcpy (command, entry->command, sizeof (command));
     
-    if (rfc1524_expand_filename (entry->nametemplate, a->filename,
+    if (fp)
+    {
+      fname = safe_strdup (a->filename);
+      mutt_sanitize_filename (fname, 1);
+    }
+    else
+      fname = a->filename;
+
+    if (rfc1524_expand_filename (entry->nametemplate, fname,
 				 tempfile, sizeof (tempfile)))
     {
-      if (fp == NULL)
+      if (fp == NULL && mutt_strcmp(tempfile, a->filename))
       {
 	/* send case: the file is already there */
-	if (mutt_rename_file (a->filename, tempfile))
+	if (safe_symlink (a->filename, tempfile) == -1)
 	{
 	  if (mutt_yesorno ("Can't match nametemplate, continue?", 1) == M_YES)
 	    strfcpy (tempfile, a->filename, sizeof (tempfile));
@@ -294,19 +437,17 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
 	    goto return_error;
 	}
 	else
-	{
-	  safe_free ((void **) &a->filename);
-	  a->filename = safe_strdup (tempfile);
-	}
+	  unlink_tempfile = 1;
       }
     }
     else if (fp == NULL) /* send case */
       strfcpy (tempfile, a->filename, sizeof (tempfile));
-
+    
     if (fp)
     {
       /* recv case: we need to save the attachment to a file */
-      if (mutt_save_attachment (fp, a, tempfile, 0) == -1)
+      FREE (&fname);
+      if (mutt_save_attachment (fp, a, tempfile, 0, NULL) == -1)
 	goto return_error;
     }
 
@@ -321,7 +462,7 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
     {
       /* recv case */
       strfcpy (pagerfile, a->filename, sizeof (pagerfile));
-      mutt_adv_mktemp (pagerfile);
+      mutt_adv_mktemp (pagerfile, sizeof(pagerfile));
     }
     else
       mutt_mktemp (pagerfile);
@@ -330,51 +471,41 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
   if (use_mailcap)
   {
     pid_t thepid = 0;
-    FILE *pagerfp = NULL;
-    FILE *tempfp = NULL;
-    FILE *filter_in;
-    FILE *filter_out;
-
+    int tempfd = -1, pagerfd = -1;
+    
     if (!use_pager)
       endwin ();
 
     if (use_pager || use_pipe)
     {
-      if (use_pager && ((pagerfp = safe_fopen (pagerfile, "w")) == NULL))
+      if (use_pager && ((pagerfd = safe_open (pagerfile, O_CREAT | O_EXCL | O_WRONLY)) == -1))
       {
-	mutt_perror ("fopen");
+	mutt_perror ("open");
 	goto return_error;
       }
-      if (use_pipe && ((tempfp = fopen (tempfile, "r")) == NULL))
+      if (use_pipe && ((tempfd = open (tempfile, 0)) == -1))
       {
-	if (pagerfp)
-	  fclose (pagerfp);
-	mutt_perror ("fopen");
+	if(pagerfd != -1)
+	  close(pagerfd);
+	mutt_perror ("open");
 	goto return_error;
       }
 
-      if ((thepid = mutt_create_filter (command, use_pipe ? &filter_in : NULL,
-					use_pager ? &filter_out : NULL, NULL)) == -1)
+      if ((thepid = mutt_create_filter_fd (command, NULL, NULL, NULL,
+					   use_pipe ? tempfd : -1, use_pager ? pagerfd : -1, -1)) == -1)
       {
-	if (pagerfp)
-	  fclose (pagerfp);
-	if (tempfp)
-	  fclose (tempfp);
+	if(pagerfd != -1)
+	  close(pagerfd);
+	
+	if(tempfd != -1)
+	  close(tempfd);
+
 	mutt_error ("Cannot create filter");
 	goto return_error;
       }
 
-      if (use_pipe)
-      {
-	mutt_copy_stream (tempfp, filter_in);
-	fclose (tempfp);
-	fclose (filter_in);
-      }
       if (use_pager)
       {
-	mutt_copy_stream (filter_out, pagerfp);
-	fclose (filter_out);
-	fclose (pagerfp);
 	if (a->description)
 	  snprintf (descrip, sizeof (descrip),
 		    "---Command: %-20.20s Description: %s",
@@ -387,6 +518,10 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
       if ((mutt_wait_filter (thepid) || (entry->needsterminal &&
 	  option (OPTWAITKEY))) && !use_pager)
 	mutt_any_key_to_continue (NULL);
+      
+      close(tempfd);
+      close(pagerfd);
+      
     }
     else
     {
@@ -403,7 +538,7 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
     if (flag == M_AS_TEXT)
     {
       /* just let me see the raw data */
-      if (mutt_save_attachment (fp, a, pagerfile, 0))
+      if (mutt_save_attachment (fp, a, pagerfile, 0, NULL))
 	goto return_error;
     }
     else
@@ -438,7 +573,8 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
     info.fp = fp;
     info.bdy = a;
     info.ctx = Context;
-    rc = mutt_do_pager (descrip, pagerfile, is_message, &info);
+    rc = mutt_do_pager (descrip, pagerfile,
+			is_message ? M_PAGER_MESSAGE : 0, &info);
   }
   else
     rc = 0;
@@ -449,6 +585,9 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
     rfc1524_free_entry (&entry);
   if (fp && tempfile[0])
     mutt_unlink (tempfile);
+  else if (unlink_tempfile)
+    unlink(tempfile);
+
   if (pagerfile[0])
     mutt_unlink (pagerfile);
 
@@ -458,15 +597,13 @@ int mutt_view_attachment (FILE *fp, BODY *a, int flag)
 /* returns 1 on success, 0 on error */
 int mutt_pipe_attachment (FILE *fp, BODY *b, const char *path, char *outfile)
 {
-  STATE o;
   pid_t thepid;
-
-  memset (&o, 0, sizeof (STATE));
+  int out = -1;
 
   if (outfile && *outfile)
-    if ((o.fpout = safe_fopen (outfile, "w")) == NULL)
+    if ((out = safe_open (outfile, O_CREAT | O_EXCL | O_WRONLY)) < 0)
     {
-      mutt_perror ("fopen");
+      mutt_perror ("open");
       return 0;
     }
 
@@ -481,7 +618,7 @@ int mutt_pipe_attachment (FILE *fp, BODY *b, const char *path, char *outfile)
     memset (&s, 0, sizeof (STATE));
 
     if (outfile && *outfile)
-      thepid = mutt_create_filter (path, &s.fpout, &o.fpin, NULL);
+      thepid = mutt_create_filter_fd (path, &s.fpout, NULL, NULL, -1, out, -1);
     else
       thepid = mutt_create_filter (path, &s.fpout, NULL, NULL);
 
@@ -498,12 +635,16 @@ int mutt_pipe_attachment (FILE *fp, BODY *b, const char *path, char *outfile)
     if ((ifp = fopen (b->filename, "r")) == NULL)
     {
       mutt_perror ("fopen");
-      fclose (o.fpout);
+      if (outfile && *outfile)
+      {
+	close (out);
+	unlink (outfile);
+      }
       return 0;
     }
 
     if (outfile && *outfile)
-      thepid = mutt_create_filter (path, &ofp, &o.fpin, NULL);
+      thepid = mutt_create_filter_fd (path, &ofp, NULL, NULL, -1, out, -1);
     else
       thepid = mutt_create_filter (path, &ofp, NULL, NULL);
 
@@ -513,11 +654,7 @@ int mutt_pipe_attachment (FILE *fp, BODY *b, const char *path, char *outfile)
   }
 
   if (outfile && *outfile)
-  {
-    mutt_copy_stream (o.fpin, o.fpout);
-    fclose (o.fpin);
-    fclose (o.fpout);
-  }
+    close (out);
 
   if (mutt_wait_filter (thepid) != 0 || option (OPTWAITKEY))
     mutt_any_key_to_continue (NULL);
@@ -525,31 +662,79 @@ int mutt_pipe_attachment (FILE *fp, BODY *b, const char *path, char *outfile)
 }
 
 /* returns 0 on success, -1 on error */
-int mutt_save_attachment (FILE *fp, BODY *m, char *path, int flags)
+int mutt_save_attachment (FILE *fp, BODY *m, char *path, int flags, HEADER *hdr)
 {
   if (fp)
   {
-    /* In recv mode, extract from folder and decode */
+    
+    /* recv mode */
 
-    STATE s;
-  
-    memset (&s, 0, sizeof (s));
-    if (flags == M_SAVE_APPEND)
-      s.fpout = safe_fopen (path, "a");
-    else
-      s.fpout = fopen (path, "w");
-    if (s.fpout == NULL)
+    if(hdr &&
+	m->hdr &&
+	m->encoding != ENCBASE64 &&
+	m->encoding != ENCQUOTEDPRINTABLE &&
+	mutt_is_message_type(m->type, m->subtype))
     {
-      mutt_perror ("fopen");
-      return (-1);
+      /* message type attachments are written to mail folders. */
+
+      char buf[HUGE_STRING];
+      HEADER *hn;
+      CONTEXT ctx;
+      MESSAGE *msg;
+      int chflags = 0;
+      int r = -1;
+      
+      hn = m->hdr;
+      hn->msgno = hdr->msgno; /* required for MH/maildir */
+      hn->read = 1;
+
+      fseek (fp, m->offset, 0);
+      if (fgets (buf, sizeof (buf), fp) == NULL)
+	return -1;
+      if (mx_open_mailbox(path, M_APPEND | M_QUIET, &ctx) == NULL)
+	return -1;
+      if ((msg = mx_open_new_message (&ctx, hn, is_from (buf, NULL, 0) ? 0 : M_ADD_FROM)) == NULL)
+      {
+	mx_close_mailbox(&ctx);
+	return -1;
+      }
+      if (ctx.magic == M_MBOX || ctx.magic == M_MMDF || ctx.magic == M_KENDRA)
+	chflags = CH_FROM;
+      chflags |= (ctx.magic == M_MAILDIR ? CH_NOSTATUS : CH_UPDATE);
+      if ((r = _mutt_copy_message (msg->fp, fp, hn, hn->content, 0, chflags)) == 0)
+	mutt_message ("Attachment saved.");
+
+      mx_commit_message (msg, &ctx);
+      mx_close_message (&msg);
+      mx_close_mailbox (&ctx);
+      return r;
     }
-    fseek ((s.fpin = fp), m->offset, 0);
-    mutt_decode_attachment (m, &s);
-
-    if (fclose (s.fpout) != 0)
+    else
     {
-      mutt_perror ("fclose");
-      return (-1);
+      /* In recv mode, extract from folder and decode */
+      
+      STATE s;
+      
+      memset (&s, 0, sizeof (s));
+      if (flags == M_SAVE_APPEND)
+	s.fpout = fopen (path, "a");
+      else if (flags == M_SAVE_OVERWRITE)
+	s.fpout = fopen (path, "w");
+      else
+	s.fpout = safe_fopen (path, "w");
+      if (s.fpout == NULL)
+      {
+	mutt_perror ("fopen");
+	return (-1);
+      }
+      fseek ((s.fpin = fp), m->offset, 0);
+      mutt_decode_attachment (m, &s);
+      
+      if (fclose (s.fpout) != 0)
+      {
+	mutt_perror ("fclose");
+	return (-1);
+      }
     }
   }
   else
@@ -591,14 +776,21 @@ int mutt_decode_save_attachment (FILE *fp, BODY *m, char *path,
 {
   STATE s;
   unsigned int saved_encoding = 0;
+  BODY *saved_parts = NULL;
+  HEADER *saved_hdr = NULL;
 
   memset (&s, 0, sizeof (s));
   s.flags = displaying ? M_DISPLAY : 0;
 
+  s.flags |= M_CHARCONV;
+  
   if (flags == M_SAVE_APPEND)
-    s.fpout = safe_fopen (path, "a");
-  else
+    s.fpout = fopen (path, "a");
+  else if (flags == M_SAVE_OVERWRITE)
     s.fpout = fopen (path, "w");
+  else
+    s.fpout = safe_fopen (path, "w");
+
   if (s.fpout == NULL)
   {
     perror ("fopen");
@@ -628,10 +820,9 @@ int mutt_decode_save_attachment (FILE *fp, BODY *m, char *path,
     m->length = st.st_size;
     m->encoding = ENC8BIT;
     m->offset = 0;
-    if (m->type == TYPEMESSAGE && m->subtype &&
-		  (!strcasecmp (m->subtype,"rfc822") ||
-		   !strcasecmp (m->subtype, "news")))
-      m->parts = mutt_parse_messageRFC822 (s.fpin, m);
+    saved_parts = m->parts;
+    saved_hdr = m->hdr;
+    mutt_parse_part (s.fpin, m);
   }
   else
     s.fpin = fp;
@@ -643,8 +834,12 @@ int mutt_decode_save_attachment (FILE *fp, BODY *m, char *path,
   {
     m->length = 0;
     m->encoding = saved_encoding;
-    if (m->parts)
-      mutt_free_body (&m->parts);
+    if (saved_parts)
+    {
+      mutt_free_header (&m->hdr);
+      m->parts = saved_parts;
+      m->hdr = saved_hdr;
+    }
     fclose (s.fpin);
   }
 
@@ -664,14 +859,15 @@ int mutt_print_attachment (FILE *fp, BODY *a)
   char type[STRING];
   pid_t thepid;
   FILE *ifp, *fpout;
-
-  snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);
+  short unlink_newfile = 0;
+  
+  snprintf (type, sizeof (type), "%s/%s", TYPE (a), a->subtype);
 
   if (rfc1524_mailcap_lookup (a, type, NULL, M_PRINT)) 
   {
     char command[_POSIX_PATH_MAX+STRING];
     rfc1524_entry *entry;
-    int piped = FALSE;
+    int piped = MUTT_FALSE;
 
     entry = rfc1524_new_entry ();
     rfc1524_mailcap_lookup (a, type, entry, M_PRINT);
@@ -680,9 +876,7 @@ int mutt_print_attachment (FILE *fp, BODY *a)
     {
       if (!fp)
       {
-	/* only attempt file move in send mode */
-
-	if (mutt_rename_file (a->filename, newfile))
+	if (safe_symlink(a->filename, newfile) == -1)
 	{
 	  if (mutt_yesorno ("Can't match nametemplate, continue?", 1) != M_YES)
 	  {
@@ -692,16 +886,13 @@ int mutt_print_attachment (FILE *fp, BODY *a)
 	  strfcpy (newfile, a->filename, sizeof (newfile));
 	}
 	else
-	{
-	  safe_free ((void **)&a->filename);
-	  a->filename = safe_strdup (newfile);
-	}
+	  unlink_newfile = 1;
       }
     }
 
     /* in recv mode, save file to newfile first */
     if (fp)
-      mutt_save_attachment (fp, a, newfile, 0);
+      mutt_save_attachment (fp, a, newfile, 0, NULL);
 
     strfcpy (command, entry->printcommand, sizeof (command));
     piped = rfc1524_expand_command (a, newfile, type, command, sizeof (command));
@@ -733,14 +924,18 @@ int mutt_print_attachment (FILE *fp, BODY *a)
 
     if (fp)
       mutt_unlink (newfile);
+    else if (unlink_newfile)
+      unlink(newfile);
 
     rfc1524_free_entry (&entry);
     return (1);
   }
 
-  if (!strcasecmp ("text/plain", a->subtype) ||
-      !strcasecmp ("application/postscript", a->subtype))
-    return (mutt_pipe_attachment (fp, a, PrintCmd, NULL));
+  if (!mutt_strcasecmp ("text/plain", a->subtype) ||
+      !mutt_strcasecmp ("application/postscript", a->subtype))
+  {
+    return (mutt_pipe_attachment (fp, a, NONULL(PrintCmd), NULL));
+  }
   else if (mutt_can_decode (a))
   {
     /* decode and print */
@@ -753,7 +948,7 @@ int mutt_print_attachment (FILE *fp, BODY *a)
       if ((ifp = fopen (newfile, "r")) != NULL)
       {
 	endwin ();
-	thepid = mutt_create_filter (PrintCmd, &fpout, NULL, NULL);
+	thepid = mutt_create_filter (NONULL(PrintCmd), &fpout, NULL, NULL);
 	mutt_copy_stream (ifp, fpout);
 	fclose (ifp);
 	fclose (fpout);
