@@ -19,6 +19,12 @@
 #include "mutt.h"
 #include "mutt_curses.h"
 #include "mime.h"
+#include "mailbox.h"
+#include "mx.h"
+
+#ifdef _PGPPATH
+#include "pgp.h"
+#endif
 
 #include <string.h>
 #include <ctype.h>
@@ -63,6 +69,7 @@ void mutt_free_body (BODY **p)
       unlink (b->filename);
     safe_free ((void **) &b->filename);
     safe_free ((void **) &b->content);
+    safe_free ((void **) &b->xtype);
     safe_free ((void **) &b->subtype);
     safe_free ((void **) &b->description);
     safe_free ((void **) &b->form_name);
@@ -143,6 +150,7 @@ HEADER *mutt_dup_header(HEADER *h)
 
 void mutt_free_header (HEADER **h)
 {
+  if(!h || !*h) return;
   mutt_free_envelope (&(*h)->env);
   mutt_free_body (&(*h)->content);
   safe_free ((void **) &(*h)->tree);
@@ -155,7 +163,7 @@ int mutt_matches_ignore (const char *s, LIST *t)
 {
   for (; t; t = t->next)
   {
-    if (!strncasecmp (s, t->data, strlen (t->data)) || *t->data == '*')
+    if (!mutt_strncasecmp (s, t->data, mutt_strlen (t->data)) || *t->data == '*')
       return 1;
   }
   return 0;
@@ -194,7 +202,7 @@ char *mutt_expand_path (char *s, size_t slen)
   if (*s == '~')
   {
     if (*(s + 1) == '/' || *(s + 1) == 0)
-      snprintf (p, sizeof (p), "%s%s", Homedir, s + 1);
+      snprintf (p, sizeof (p), "%s%s", NONULL(Homedir), s + 1);
     else
     {
       struct passwd *pw;
@@ -267,14 +275,14 @@ void *safe_calloc (size_t nmemb, size_t size)
     return NULL;
   if (!(p = calloc (nmemb, size)))
   {
-    mutt_error ("Out of memory");
+    mutt_error ("Out of memory!");
     sleep (1);
     mutt_exit (1);
   }
   return p;
 }
 
-void *safe_malloc (unsigned int siz)
+void *safe_malloc (size_t siz)
 {
   void *p;
 
@@ -330,12 +338,24 @@ void safe_free (void **p)
   }
 }
 
+int safe_fclose (FILE **f)
+{
+  int r = 0;
+  
+  if (*f)
+    r = fclose (*f);
+      
+  *f = NULL;
+  return r;
+}
+
 char *safe_strdup (const char *s)
 {
   char *p;
   size_t l;
 
-  if (!s || !*s) return 0;
+  if (!s || !*s)
+    return 0;
   l = strlen (s) + 1;
   p = (char *)safe_malloc (l);
   memcpy (p, s, l);
@@ -372,12 +392,35 @@ int mutt_copy_bytes (FILE *in, FILE *out, size_t size)
 char *mutt_get_parameter (const char *s, PARAMETER *p)
 {
   for (; p; p = p->next)
-    if (strcasecmp (s, p->attribute) == 0)
+    if (mutt_strcasecmp (s, p->attribute) == 0)
       return (p->value);
 
   return NULL;
 }
 
+void mutt_set_parameter (const char *attribute, const char *value, PARAMETER **p)
+{
+  PARAMETER *q;
+  for(q = *p; q; q =  q->next)
+  {
+    if (mutt_strcasecmp (attribute, q->attribute) == 0)
+    {
+      safe_free((void **) &q->value);
+      q->value = safe_strdup(value);
+      return;
+    }
+  }
+  
+  q = mutt_new_parameter();
+  q->attribute = safe_strdup(attribute);
+  q->value = safe_strdup(value);
+  q->next = *p;
+  *p = q;
+}
+
+
+
+
 /* returns 1 if Mutt can't display this type of data, 0 otherwise */
 int mutt_needs_mailcap (BODY *m)
 {
@@ -385,15 +428,20 @@ int mutt_needs_mailcap (BODY *m)
   {
     case TYPETEXT:
 
-      if (!strcasecmp ("plain", m->subtype) ||
-	  !strcasecmp ("rfc822-headers", m->subtype) ||
-	  !strcasecmp ("enriched", m->subtype))
+      if (!mutt_strcasecmp ("plain", m->subtype) ||
+	  !mutt_strcasecmp ("rfc822-headers", m->subtype) ||
+	  !mutt_strcasecmp ("enriched", m->subtype))
 	return 0;
       break;
 
 
 
-
+#ifdef _PGPPATH
+    case TYPEAPPLICATION:
+      if(mutt_is_application_pgp(m))
+	return 0;
+      break;
+#endif /* _PGPPATH */
 
 
     case TYPEMULTIPART:
@@ -412,12 +460,19 @@ int mutt_is_text_type (int t, char *s)
 
   if (t == TYPEMESSAGE)
   {
-    if (!strcasecmp ("delivery-status", s))
+    if (!mutt_strcasecmp ("delivery-status", s))
       return 1;
   }
 
 
 
+#ifdef _PGPPATH
+  if (t == TYPEAPPLICATION)
+  {
+    if (!mutt_strcasecmp ("pgp-keys", s))
+      return 1;
+  }
+#endif /* _PGPPATH */
 
 
 
@@ -437,6 +492,8 @@ void mutt_free_envelope (ENVELOPE **p)
   rfc822_free_address (&(*p)->mail_followup_to);
   safe_free ((void **) &(*p)->subject);
   safe_free ((void **) &(*p)->message_id);
+  safe_free ((void **) &(*p)->supersedes);
+  safe_free ((void **) &(*p)->date);
   mutt_free_list (&(*p)->references);
   mutt_free_list (&(*p)->userhdrs);
   safe_free ((void **) p);
@@ -454,7 +511,7 @@ void mutt_tabs_to_spaces (char *s)
 
 void mutt_mktemp (char *s)
 {
-  snprintf (s, _POSIX_PATH_MAX, "%s/mutt-%s-%d-%d", NONULL (Tempdir), Hostname, (int) getpid (), Counter++);
+  snprintf (s, _POSIX_PATH_MAX, "%s/mutt-%s-%d-%d", NONULL (Tempdir), NONULL(Hostname), (int) getpid (), Counter++);
   unlink (s);
 }
 
@@ -462,6 +519,8 @@ void mutt_perror (const char *s)
 {
   const char *p = strerror (errno);
 
+  dprint (1, (debugfile, "%s: %s (errno = %d)\n", s, 
+      p ? p : "unknown error", errno));
   mutt_error ("%s: %s (errno = %d)", s, p ? p : "unknown error", errno);
 }
 
@@ -479,20 +538,6 @@ char *mutt_strlower (char *s)
   return (s);
 }
 
-/* strcmp() allowing NULL pointers */
-int mutt_strcmp (const char *s1, const char *s2)
-{
-  if (s1 != NULL)
-  {
-    if (s2 != NULL)
-      return strcmp (s1, s2);
-    else
-      return (1);
-  }
-  else
-    return ((s2 == NULL) ? 0 : -1);
-}
-
 #ifndef LIBMUTT
 void mutt_free_alias (ALIAS **p)
 {
@@ -504,7 +549,7 @@ void mutt_free_alias (ALIAS **p)
     *p = (*p)->next;
     safe_free ((void **) &t->name);
     rfc822_free_address (&t->addr);
-    free (t);
+    safe_free ((void **) &t);
   }
 }
 #endif
@@ -532,12 +577,12 @@ void mutt_pretty_mailbox (char *s)
   }
   *q = 0;
 
-  if (strncmp (s, NONULL (Maildir), (len = strlen (NONULL (Maildir)))) == 0 && s[len] == '/')
+  if (mutt_strncmp (s, Maildir, (len = mutt_strlen (Maildir))) == 0 && s[len] == '/')
   {
     *s++ = '=';
     strcpy (s, s + len);
   }
-  else if (strncmp (s, Homedir, (len = strlen (Homedir))) == 0 &&
+  else if (mutt_strncmp (s, Homedir, (len = mutt_strlen (Homedir))) == 0 &&
 	   s[len] == '/')
   {
     *s++ = '~';
@@ -581,14 +626,24 @@ int mutt_copy_stream (FILE *fin, FILE *fout)
   return 0;
 }
 
+void mutt_expand_file_fmt (char *dest, size_t destlen, const char *fmt, const char *src)
+{
+  char *_src = mutt_quote_filename(src);
+  
+  mutt_expand_fmt(dest, destlen, fmt, _src);
+  safe_free((void **) &_src);
+}
+
 void mutt_expand_fmt (char *dest, size_t destlen, const char *fmt, const char *src)
 {
   const char *p = fmt;
   const char *last = p;
   size_t len;
-  size_t slen = strlen (src);
+  size_t slen;
   int found = 0;
 
+  slen = mutt_strlen (src);
+  
   while ((p = strchr (p, '%')) != NULL)
   {
     if (p[1] == 's')
@@ -633,7 +688,132 @@ void mutt_expand_fmt (char *dest, size_t destlen, const char *fmt, const char *s
   if (found)
     strfcpy (dest, last, destlen);
   else
-    snprintf (dest, destlen, "%s '%s'", fmt, src);
+    snprintf (dest, destlen, "%s %s", fmt, src);
+  
+}
+
+static int 
+compare_stat (struct stat *osb, struct stat *nsb)
+{
+  if (osb->st_dev != nsb->st_dev || osb->st_ino != nsb->st_ino ||
+      osb->st_rdev != nsb->st_rdev)
+  {
+    return -1;
+  }
+
+  return 0;
+}
+
+int safe_symlink(const char *oldpath, const char *newpath)
+{
+  struct stat osb, nsb;
+
+  if(!oldpath || !newpath)
+    return -1;
+  
+  if(unlink(newpath) == -1 && errno != ENOENT)
+    return -1;
+  
+  if (oldpath[0] == '/')
+  {
+    if (symlink (oldpath, newpath) == -1)
+      return -1;
+  }
+  else
+  {
+    char abs_oldpath[_POSIX_PATH_MAX];
+
+    if ((getcwd (abs_oldpath, sizeof abs_oldpath) == NULL) ||
+	(strlen (abs_oldpath) + 1 + strlen (oldpath) + 1 > sizeof abs_oldpath))
+    return -1;
+  
+    strcat (abs_oldpath, "/");
+    strcat (abs_oldpath, oldpath);
+    if (symlink (abs_oldpath, newpath) == -1)
+      return -1;
+  }
+
+  if(stat(oldpath, &osb) == -1 || stat(newpath, &nsb) == -1
+     || compare_stat(&osb, &nsb) == -1)
+  {
+    unlink(newpath);
+    return -1;
+  }
+  
+  return 0;
+}
+
+/* 
+ * This function is supposed to do nfs-safe renaming of files.
+ * 
+ * Warning: We don't check whether src and target are equal.
+ */
+
+int safe_rename (const char *src, const char *target)
+{
+  struct stat ssb, tsb;
+
+  if (!src || !target)
+    return -1;
+
+  if (link (src, target) != 0)
+  {
+    return -1;
+  }
+
+  /*
+   * Stat both links and check if they are equal.
+   */
+  
+  if (stat (src, &ssb) == -1)
+  {
+    return -1;
+  }
+  
+  if (stat (target, &tsb) == -1)
+  {
+    return -1;
+  }
+
+  /* 
+   * pretend that the link failed because the target file
+   * did already exist.
+   */
+
+  if (compare_stat (&ssb, &tsb) == -1)
+  {
+    errno = EEXIST;
+    return -1;
+  }
+
+  /*
+   * Unlink the original link.  Should we really ignore the return
+   * value here? XXX
+   */
+
+  unlink (src);
+
+  return 0;
+}
+
+int safe_open (const char *path, int flags)
+{
+  struct stat osb, nsb;
+  int fd;
+
+  if ((fd = open (path, flags, 0600)) < 0)
+    return fd;
+
+  /* make sure the file is not symlink */
+  if (lstat (path, &osb) < 0 || fstat (fd, &nsb) < 0 ||
+      compare_stat(&osb, &nsb) == -1)
+  {
+    dprint (1, (debugfile, "safe_open(): %s is a symlink!\n", path));
+    close (fd);
+    return (-1);
+  }
+
+  return (fd);
 }
 
 /* when opening files for writing, make sure the file doesn't already exist
@@ -641,8 +821,6 @@ void mutt_expand_fmt (char *dest, size_t destlen, const char *fmt, const char *s
  */
 FILE *safe_fopen (const char *path, const char *mode)
 {
-  struct stat osb, nsb;
-
   if (mode[0] == 'w')
   {
     int fd;
@@ -653,18 +831,8 @@ FILE *safe_fopen (const char *path, const char *mode)
     else
       flags |= O_WRONLY;
 
-    if ((fd = open (path, flags, 0600)) < 0)
-      return NULL;
-
-    /* make sure the file is not symlink */
-    if (lstat (path, &osb) < 0 || fstat (fd, &nsb) < 0 ||
-	osb.st_dev != nsb.st_dev || osb.st_ino != nsb.st_ino ||
-	osb.st_rdev != nsb.st_rdev)
-    {
-      dprint (1, (debugfile, "safe_fopen():%s is a symlink!\n", path));
-      close (fd);
+    if ((fd = safe_open (path, flags)) < 0)
       return (NULL);
-    }
 
     return (fdopen (fd, mode));
   }
@@ -674,7 +842,7 @@ FILE *safe_fopen (const char *path, const char *mode)
 #ifndef LIBMUTT
 /* return 0 on success, -1 on error */
 int mutt_check_overwrite (const char *attname, const char *path,
-				char *fname, size_t flen, int flags) 
+				char *fname, size_t flen, int *append) 
 {
   char tmp[_POSIX_PATH_MAX];
   struct stat st;
@@ -699,12 +867,24 @@ int mutt_check_overwrite (const char *attname, const char *path,
     else
       snprintf (fname, flen, "%s/%s", path, attname);
   }
-
-  if (flags != M_SAVE_APPEND &&
-      access (fname, F_OK) == 0 && 
-      mutt_yesorno ("File exists, overwrite?", 0) != 1)
-    return (-1);
   
+  if (*append == 0 && access (fname, F_OK) == 0)
+  {
+    switch (mutt_multi_choice
+	    ("File exists, (o)verwrite, (a)ppend, or (c)ancel?", "oac"))
+    {
+      case -1: /* abort */
+      case 3:  /* cancel */
+	return -1;
+
+      case 2: /* append */
+        *append = M_SAVE_APPEND;
+        break;
+      case 1: /* overwrite */
+        *append = M_SAVE_OVERWRITE;
+        break;
+    }
+  }
   return 0;
 }
 #endif
@@ -712,7 +892,7 @@ void mutt_remove_trailing_ws (char *s)
 {
   char *p;
 
-  for (p = s + strlen (s) - 1 ; p >= s && ISSPACE (*p) ; p--)
+  for (p = s + mutt_strlen (s) - 1 ; p >= s && ISSPACE (*p) ; p--)
     *p = 0;
 }
 
@@ -766,6 +946,19 @@ void mutt_safe_path (char *s, size_t l, ADDRESS *a)
       *p = '_';
 }
 
+static char safe_chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+@{}._-:%/";
+
+void mutt_sanitize_filename(char *f, short slash)
+{
+  if(!f) return;
+
+  for(; *f; f++)
+  {
+    if((slash && *f == '/') || !strchr(safe_chars, *f))
+      *f = '_';
+  }
+}
+
 /* Read a line from ``fp'' into the dynamically allocated ``s'',
  * increasing ``s'' if necessary. The ending "\n" or "\r\n" is removed.
  * If a line ends with "\", this char and the linefeed is removed,
@@ -786,7 +979,7 @@ char *mutt_read_line (char *s, size_t *size, FILE *fp, int *line)
   {
     if (fgets (s + offset, *size - offset, fp) == NULL)
     {
-      free (s);
+      safe_free ((void **) &s);
       return NULL;
     }
     if ((ch = strchr (s + offset, '\n')) != NULL)
@@ -837,16 +1030,17 @@ char *mutt_substrdup (const char *begin, const char *end)
 void mutt_FormatString (char *dest,		/* output buffer */
 			size_t destlen,		/* output buffer len */
 			const char *src,	/* template string */
-			const int width,	/* width to format to */
 			format_t *callback,	/* callback for processing */
 			unsigned long data,	/* callback data */
 			format_flag flags)	/* callback flags */
 {
   char prefix[SHORT_STRING], buf[LONG_STRING], *cp, *wptr = dest, ch;
   char ifstring[SHORT_STRING], elsestring[SHORT_STRING];
-  size_t wlen = 0, count, len;
+  size_t wlen, count, len;
 
   destlen--; /* save room for the terminal \0 */
+  wlen = (flags & M_FORMAT_ARROWCURSOR && option (OPTARROWCURSOR)) ? 3 : 0;
+    
   while (*src && wlen < destlen)
   {
     if (*src == '%')
@@ -931,8 +1125,8 @@ void mutt_FormatString (char *dest,		/* output buffer */
 	if (count > wlen)
 	{
 	  count -= wlen; /* how many chars left on this line */
-	  mutt_FormatString (buf, sizeof (buf), src, COLS, callback, data, flags);
-	  len = strlen (buf);
+	  mutt_FormatString (buf, sizeof (buf), src, callback, data, flags);
+	  len = mutt_strlen (buf);
 	  if (count > len)
 	  {
 	    count -= len; /* how many chars to pad */
@@ -970,11 +1164,8 @@ void mutt_FormatString (char *dest,		/* output buffer */
 	/* use callback function to handle this case */
 	src = callback (buf, sizeof (buf), ch, src, prefix, ifstring, elsestring, data, flags);
 
-	if ((len = strlen (buf)) + wlen > destlen)
-	{
-	  if ((len = destlen - wlen) < 0)
-	    len = 0;
-	}
+	if ((len = mutt_strlen (buf)) + wlen > destlen)
+	  len = (destlen - wlen > 0) ? (destlen - wlen) : 0;
 	memcpy (wptr, buf, len);
 	wptr += len;
 	wlen += len;
@@ -1034,7 +1225,7 @@ void mutt_FormatString (char *dest,		/* output buffer */
 FILE *mutt_open_read (const char *path, pid_t *thepid)
 {
   FILE *f;
-  int len = strlen (path);
+  int len = mutt_strlen (path);
 
   if (path[len - 1] == '|')
   {
@@ -1047,7 +1238,7 @@ FILE *mutt_open_read (const char *path, pid_t *thepid)
     endwin ();
 #endif    
     *thepid = mutt_create_filter (s, NULL, &f, NULL);
-    free (s);
+    safe_free ((void **) &s);
   }
   else
   {
@@ -1056,3 +1247,156 @@ FILE *mutt_open_read (const char *path, pid_t *thepid)
   }
   return (f);
 }
+
+/* returns 1 if OK to proceed, 0 to abort */
+int mutt_save_confirm (const char *s, struct stat *st)
+{
+#ifndef LIBMUTT
+  char tmp[_POSIX_PATH_MAX];
+  int ret = 1;
+  int magic = 0;
+
+  magic = mx_get_magic (s);
+
+  if (stat (s, st) != -1)
+  {
+    if (magic == -1)
+    {
+      mutt_error ("%s is not a mailbox!", s);
+      return 0;
+    }
+
+    if (option (OPTCONFIRMAPPEND))
+    {
+      snprintf (tmp, sizeof (tmp), "Append messages to %s?", s);
+      if (mutt_yesorno (tmp, 1) < 1)
+	ret = 0;
+    }
+  }
+  else
+  {
+#ifdef USE_IMAP
+    if (magic != M_IMAP)
+#endif /* execute the block unconditionally if we don't use imap */
+    {
+      st->st_mtime = 0;
+      st->st_atime = 0;
+
+      if (errno == ENOENT)
+      {
+	if (option (OPTCONFIRMCREATE))
+	{
+	  snprintf (tmp, sizeof (tmp), "Create %s?", s);
+	  if (mutt_yesorno (tmp, 1) < 1)
+	    ret = 0;
+	}
+      }
+      else
+      {
+	mutt_perror (s);
+	return 0;
+      }
+    }
+  }
+
+  CLEARLINE (LINES-1);
+  return (ret);
+#else
+  return 1;
+#endif
+}
+
+/* prepare a file name to survive the shell's quoting rules.
+ * From the Unix programming FAQ by way of Liviu.
+ */
+
+char *mutt_quote_filename(const char *f)
+{
+  char *d;
+  size_t i,l;
+
+  if(!f) return NULL;
+  
+  for(i = 0, l = 3; f[i]; i++, l++)
+  {
+    if(f[i] == '\'' || f[i] == '`')
+      l += 3;
+  }
+  
+  d = safe_malloc(l);
+  
+  l = 0;
+  d[l++] = '\'';
+  
+  for(i = 0; f[i]; i++)
+  {
+    if(f[i] == '\'' || f[i] == '`')
+    {
+      d[l++] = '\'';
+      d[l++] = '\\';
+      d[l++] = f[i];
+      d[l++] = '\'';
+    }
+    else
+      d[l++] = f[i];
+  }
+  
+  d[l++] = '\'';
+  d[l]   = '\0';
+  
+  return d;
+}
+
+void state_prefix_putc(char c, STATE *s)
+{
+  if (s->flags & M_PENDINGPREFIX)
+  {
+    state_reset_prefix(s);
+    if (s->prefix)
+      state_puts(s->prefix, s);
+  }
+
+  state_putc(c, s);
+
+  if(c == '\n')
+    state_set_prefix(s);
+}
+
+int state_printf(STATE *s, const char *fmt, ...)
+{
+  int rv;
+  va_list ap;
+
+  va_start (ap, fmt);
+  rv = vfprintf(s->fpout, fmt, ap);
+  va_end(ap);
+  
+  return rv;
+}
+
+/* NULL-pointer aware string comparison functions */
+
+int mutt_strcmp(const char *a, const char *b)
+{
+  return strcmp(NONULL(a), NONULL(b));
+}
+
+int mutt_strcasecmp(const char *a, const char *b)
+{
+  return strcasecmp(NONULL(a), NONULL(b));
+}
+
+int mutt_strncmp(const char *a, const char *b, size_t l)
+{
+  return strncmp(NONULL(a), NONULL(b), l);
+}
+
+int mutt_strncasecmp(const char *a, const char *b, size_t l)
+{
+  return strncasecmp(NONULL(a), NONULL(b), l);
+}
+
+size_t mutt_strlen(const char *a)
+{
+  return a ? strlen (a) : 0;
+}
