@@ -1,5 +1,5 @@
 /* Balsa E-Mail Client
- * Copyright (C) 1997-98 Jay Painter and Stuart Parmenter
+ * Copyright (C) 1997-1999 Jay Painter and Stuart Parmenter
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,8 +19,8 @@
 #ifndef __NEW_MAILBOX_H__
 #define __NEW_MAILBOX_H__
 
-#include "mailbox.h"
+#include "libbalsa.h"
 
-void mailbox_conf_new       (Mailbox * mailbox, gint add_new_mailbox);
-void mailbox_conf_delete    (Mailbox * mailbox);
+void mailbox_conf_new (Mailbox * mailbox, gint add_new_mailbox, MailboxType type);
+void mailbox_conf_delete (Mailbox * mailbox);
 #endif /* __NEW_MAILBOX_H__ */
