@@ -1,5 +1,5 @@
 /* Balsa E-Mail Client
- * Copyright (C) 1997-98 Jay Painter and Stuart Parmenter
+ * Copyright (C) 1997-1999 Jay Painter and Stuart Parmenter
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,16 +16,28 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
  * 02111-1307, USA.
  */
-#ifndef __MBLIST_WINDOW_H__
-#define __MBLIST_WINDOW_H__
 
-#include "mailbox.h"
-#include "misc.h"
+#ifndef MBLIST_WINDOW_H
+#define MBLIST_WINDOW_H 1
 
-void mblist_open_window (GnomeMDI *);
-void mblist_add_mailbox (Mailbox *);
-void mblist_remove_mailbox (Mailbox * mailbox);
-void mblist_close_mailbox(Mailbox* mailbox);
-void mblist_redraw(void);
 
-#endif /* __MBLIST_WINDOW_H__ */
+#include "config.h"
+
+
+#include <gnome.h>
+#include "main-window.h"
+
+
+void mblist_menu_add_cb (GtkWidget * widget, gpointer data);
+void mblist_menu_edit_cb (GtkWidget * widget, gpointer data);
+void mblist_menu_delete_cb (GtkWidget * widget, gpointer data);
+void mblist_menu_open_cb (GtkWidget * widget, gpointer data);
+void mblist_menu_close_cb (GtkWidget * widget, gpointer data);
+
+GtkWidget *balsa_mailbox_list_window_new(BalsaWindow *window);
+
+Mailbox *mblist_get_selected_mailbox (void);
+
+
+
+#endif /* MBLIST_WINDOW_H */
