@@ -1,5 +1,5 @@
 /* Balsa E-Mail Client
- * Copyright (C) 1997-98 Jay Painter and Stuart Parmenter
+ * Copyright (C) 1997-1999 Jay Painter and Stuart Parmenter
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,12 +21,14 @@
 
 
 /* open the preferences manager window */
-void open_preferences_manager (void);
+void open_preferences_manager (GtkWidget *widget, gpointer data);
 
 /* refresh any data displayed in the preferences manager
  * window in case it has changed */
 void refresh_preferences_manager (void);
 
+/* update the pop3 server clist */
+void update_pop3_servers (void);
 
 #endif
 /* __pref_manager_h__ */
