@@ -3,7 +3,7 @@
  * This file handles Balsa's interaction with libPropList, which stores
  * Balsa's configuration information.
  *
- * This file is Copyright (C) 1998 Nat Friedman
+ * This file is Copyright (C) 1998-1999 Nat Friedman
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,6 +23,10 @@
 
 #include "config.h"
 
+#include "libbalsa.h"
+
+#include <assert.h>
+
 #include <gnome.h>
 #include <proplist.h>
 #include <sys/types.h>
@@ -30,9 +34,9 @@
 #include <fcntl.h>
 #include <string.h>
 #include "balsa-app.h"
-#include "mailbox.h"
 #include "misc.h"
 #include "save-restore.h"
+#include "../libmutt/mutt.h"
 
 static proplist_t pl_dict_add_str_str (proplist_t dict_arg, gchar * string1,
 				       gchar * string2);
@@ -48,9 +52,22 @@ rot (gchar * pass)
 {
   gchar *buff;
   gint len = 0, i = 0;
+
+  /*PKGW: let's do the assert() BEFORE we coredump... */
+  assert( pass != NULL );
+
   len = strlen (pass);
   buff = g_strdup (pass);
 
+  /* Used to assert( buff ); this is wrong. If g_strdup
+     fails we're screwed anyway to let the upcoming coredump
+     occur. */
+
+  assert (pass != NULL);
+  assert (buff != NULL);	/* TODO: using assert for this case is wrong!
+				 * The error should be handled gracefully.
+				 */
+
   for (i = 0; i < len; i++)
     {
       if ((buff[i] <= 'M' && buff[i] >= 'A')
@@ -77,11 +94,8 @@ config_load (gchar * user_filename)
 
   /* Construct the filename by appending 'user_filename' to the user's
      home directory. */
-  filename = g_malloc (strlen (user_filename) +
-		       strlen (g_get_home_dir ()) + 2);
-  strcpy (filename, g_get_home_dir ());
-  strcat (filename, "/");
-  strcat (filename, user_filename);
+
+  filename = g_strdup_printf("%s/%s", g_get_home_dir(), user_filename);
 
   balsa_app.proplist = PLGetProplistWithPath (filename);
 
@@ -113,11 +127,7 @@ config_save (gchar * user_filename)
 
   /* Construct the filename by appending 'user_filename' to the user's
      home directory. */
-  filename = g_malloc (strlen (user_filename) +
-		       strlen (g_get_home_dir ()) + 2);
-  strcpy (filename, g_get_home_dir ());
-  strcat (filename, "/");
-  strcat (filename, user_filename);
+  filename = g_strdup_printf("%s/%s", g_get_home_dir(), user_filename);
 
   /* Set the property list's filename */
   temp_str = PLMakeString (filename);
@@ -162,7 +172,7 @@ gint
 config_mailbox_add (Mailbox * mailbox, char *key_arg)
 {
   proplist_t mbox_dict, accounts, temp_str;
-  char key[32];
+  char key[MAX_PROPLIST_KEY_LEN];
 
   /* Initialize the key in case it is accidentally used uninitialized */
   strcpy (key, "AnErrorOccurred");
@@ -193,24 +203,38 @@ config_mailbox_add (Mailbox * mailbox, char *key_arg)
          Username = ...;
          Password = ...;
          Server = ...;
+	 Check = 0 | 1;
+	 Delete =  0 | 1;
        */
       mbox_dict = pl_dict_add_str_str (NULL, "Type", "POP3");
       pl_dict_add_str_str (mbox_dict, "Name", mailbox->name);
       pl_dict_add_str_str (mbox_dict, "Username",
-			   MAILBOX_POP3 (mailbox)->user);
+			   MAILBOX_POP3(mailbox)->server->user);
       /* Do not save the password if the field is NULL.  This is here
          so that asving the password to the balsarc file can be optional */
-      if (MAILBOX_POP3 (mailbox)->passwd)
+      if (MAILBOX_POP3(mailbox)->server->passwd)
 	{
 	  gchar *buff;
-	  buff = rot (MAILBOX_POP3 (mailbox)->passwd);
+	  buff = rot (MAILBOX_POP3(mailbox)->server->passwd);
 	  pl_dict_add_str_str (mbox_dict, "Password",
 			       buff);
 	  g_free (buff);
 	}
       pl_dict_add_str_str (mbox_dict, "Server",
-			   MAILBOX_POP3 (mailbox)->server);
+			   MAILBOX_POP3 (mailbox)->server->host);
+      {
+	char tmp[32];
 
+	snprintf (tmp, sizeof (tmp), "%d", MAILBOX_POP3 (mailbox)->check);
+	pl_dict_add_str_str (mbox_dict, "Check", tmp);
+
+	snprintf (tmp, sizeof (tmp), "%d", MAILBOX_POP3 (mailbox)->delete_from_server);
+	pl_dict_add_str_str (mbox_dict, "Delete", tmp);
+      }
+
+        if ((MAILBOX_POP3 (mailbox)->last_popped_uid) != NULL)
+		pl_dict_add_str_str (mbox_dict, "LastUID", MAILBOX_POP3 (mailbox)->last_popped_uid);
+	  
       break;
 
     case MAILBOX_IMAP:
@@ -226,26 +250,26 @@ config_mailbox_add (Mailbox * mailbox, char *key_arg)
       mbox_dict = pl_dict_add_str_str (NULL, "Type", "IMAP");
       pl_dict_add_str_str (mbox_dict, "Name", mailbox->name);
       pl_dict_add_str_str (mbox_dict, "Server",
-			   MAILBOX_IMAP (mailbox)->server);
+			   MAILBOX_IMAP (mailbox)->server->host);
 
       /* Add the Port entry */
       {
-	char tmp[32];
-	snprintf (tmp, sizeof (tmp), "%d", MAILBOX_IMAP (mailbox)->port);
+	char tmp[MAX_PROPLIST_KEY_LEN];
+	snprintf (tmp, sizeof (tmp), "%d", MAILBOX_IMAP(mailbox)->server->port);
 	pl_dict_add_str_str (mbox_dict, "Port", tmp);
       }
 
       pl_dict_add_str_str (mbox_dict, "Path", MAILBOX_IMAP (mailbox)->path);
       pl_dict_add_str_str (mbox_dict, "Username",
-			   MAILBOX_IMAP (mailbox)->user);
-      if (MAILBOX_IMAP (mailbox)->passwd != NULL)
-	{
-	  gchar *buff;
-	  buff = rot (MAILBOX_IMAP (mailbox)->passwd);
-	  pl_dict_add_str_str (mbox_dict, "Password",
-			       buff);
-	  g_free (buff);
-	}
+			   MAILBOX_IMAP (mailbox)->server->user);
+
+      if (MAILBOX_IMAP(mailbox)->server->passwd != NULL)
+      {
+	gchar *buff;
+	buff = rot (MAILBOX_IMAP(mailbox)->server->passwd);
+	pl_dict_add_str_str (mbox_dict, "Password", buff);
+	g_free (buff);
+      }
       break;
 
     default:
@@ -293,7 +317,7 @@ config_mailbox_add (Mailbox * mailbox, char *key_arg)
       if (key_arg == NULL)
 	{
 	  int mbox_max;
-
+	
 	  mbox_max = config_mailbox_get_highest_number (accounts);
 	  snprintf (key, sizeof (key), "m%d", mbox_max + 1);
 	}
@@ -352,18 +376,25 @@ gint
 config_mailbox_update (Mailbox * mailbox, gchar * old_mbox_name)
 {
   proplist_t mbox, mbox_key;
-  gchar *key;
+  gchar key[MAX_PROPLIST_KEY_LEN];
+
 
   mbox = config_mailbox_get_by_name (old_mbox_name);
   mbox_key = config_mailbox_get_key (mbox);
+
   if (mbox_key == NULL)
-    key = "generic";
+    {
+      strcpy (key, "generic");
+    }
   else
-    key = PLGetString (mbox_key);
+    {
+      strcpy (key, PLGetString (mbox_key));
+    }
 
   config_mailbox_delete (old_mbox_name);
   config_mailbox_add (mailbox, key);
 
+
   return config_save (BALSA_CONFIG_FILE);
 }				/* config_mailbox_update */
 
@@ -442,7 +473,7 @@ config_mailbox_init (proplist_t mbox, gchar * key)
       mailbox_type = mailbox_valid (path);
       if (mailbox_type != MAILBOX_UNKNOWN)
 	{
-	  mailbox = mailbox_new (mailbox_type);
+	  mailbox = BALSA_MAILBOX(mailbox_new (mailbox_type));
 	  mailbox->name = mailbox_name;
 	  MAILBOX_LOCAL (mailbox)->path = g_strdup (path);
 	}
@@ -455,56 +486,73 @@ config_mailbox_init (proplist_t mbox, gchar * key)
     }
   else if (!strcasecmp (type, "POP3"))	/* POP3 mailbox */
     {
-      mailbox = mailbox_new (MAILBOX_POP3);
+      mailbox = BALSA_MAILBOX(mailbox_new (MAILBOX_POP3));
       mailbox->name = mailbox_name;
 
       if ((field = pl_dict_get_str (mbox, "Username")) == NULL)
 	return FALSE;
-      MAILBOX_POP3 (mailbox)->user = g_strdup (field);
+      MAILBOX_POP3 (mailbox)->server->user = g_strdup (field);
 
       if ((field = pl_dict_get_str (mbox, "Password")) != NULL)
 	{
-	  gchar *buff;
+	  gchar *buff ;
 	  buff = rot (field);
-	  MAILBOX_POP3 (mailbox)->passwd = g_strdup (buff);
+	  MAILBOX_POP3 (mailbox)->server->passwd = g_strdup (buff);
 	  g_free (buff);
 	}
       else
-	MAILBOX_POP3 (mailbox)->passwd = NULL;
+	MAILBOX_POP3 (mailbox)->server->passwd = NULL;
 
       if ((field = pl_dict_get_str (mbox, "Server")) == NULL)
 	return FALSE;
-      MAILBOX_POP3 (mailbox)->server = g_strdup (field);
+      MAILBOX_POP3 (mailbox)->server->host = g_strdup (field);
+
+      if ((field = pl_dict_get_str (mbox, "Check")) == NULL)
+	MAILBOX_POP3 (mailbox)->check = FALSE;
+      else
+	MAILBOX_POP3 (mailbox)->check = atol (field);
+
+      if ((field = pl_dict_get_str (mbox, "Delete")) == NULL)
+	MAILBOX_POP3 (mailbox)->delete_from_server = FALSE;
+      else
+	MAILBOX_POP3 (mailbox)->delete_from_server = atol (field);
+
+	  if ((field = pl_dict_get_str (mbox, "LastUID")) == NULL)
+	MAILBOX_POP3 (mailbox)->last_popped_uid = NULL;
+      else
+	MAILBOX_POP3 (mailbox)->last_popped_uid = g_strdup (field);
+
+
 
       balsa_app.inbox_input =
 	g_list_append (balsa_app.inbox_input, mailbox);
     }
   else if (!strcasecmp (type, "IMAP"))	/* IMAP Mailbox */
     {
-      mailbox = mailbox_new (MAILBOX_IMAP);
+      mailbox = BALSA_MAILBOX(mailbox_new (MAILBOX_IMAP));
       mailbox->name = mailbox_name;
 
       if ((field = pl_dict_get_str (mbox, "Username")) == NULL)
 	return FALSE;
-      MAILBOX_IMAP (mailbox)->user = g_strdup (field);
+      MAILBOX_IMAP (mailbox)->server->user = g_strdup (field);
 
       if ((field = pl_dict_get_str (mbox, "Password")) != NULL)
-	{
-	  gchar *buff;
-	  buff = rot (field);
-	  MAILBOX_IMAP (mailbox)->passwd = g_strdup (buff);
-	  g_free (buff);
-	}
+      {
+	gchar *buff;
+	buff = rot (field);
+	MAILBOX_IMAP (mailbox)->server->passwd = g_strdup (buff);
+	g_free (buff);
+      }
       else
-	MAILBOX_IMAP (mailbox)->passwd = NULL;
+	MAILBOX_IMAP (mailbox)->server->passwd = NULL;
 
       if ((field = pl_dict_get_str (mbox, "Server")) == NULL)
 	return FALSE;
-      MAILBOX_IMAP (mailbox)->server = g_strdup (field);
+      MAILBOX_IMAP (mailbox)->server->host = g_strdup (field);
 
       if ((field = pl_dict_get_str (mbox, "Port")) == NULL)
 	return FALSE;
-      MAILBOX_IMAP (mailbox)->port = atol (field);
+      MAILBOX_IMAP (mailbox)->server->port = atol (field);
 
       if ((field = pl_dict_get_str (mbox, "Path")) == NULL)
 	return FALSE;
@@ -525,6 +573,14 @@ config_mailbox_init (proplist_t mbox, gchar * key)
     {
       balsa_app.outbox = mailbox;
     }
+  else if (strcmp ("Sentbox", key) == 0)
+    {
+      balsa_app.sentbox = mailbox;
+    }
+  else if (strcmp ("Draftbox", key) == 0)
+    {
+      balsa_app.draftbox = mailbox;
+    }
   else if (strcmp ("Trash", key) == 0)
     {
       balsa_app.trash = mailbox;
@@ -566,22 +622,97 @@ config_global_load (void)
   /* user's real name */
   if ((field = pl_dict_get_str (globals, "RealName")) == NULL)
     return FALSE;
-  balsa_app.real_name = g_strdup (field);
+  g_free (balsa_app.address->personal);
+  balsa_app.address->personal = g_strdup (field);
 
   /* user's email address */
   if ((field = pl_dict_get_str (globals, "Email")) == NULL)
     return FALSE;
-  balsa_app.email = g_strdup (field);
+  g_free (balsa_app.address->mailbox);
+  balsa_app.address->mailbox = g_strdup (field);
+
+  /* users's replyto address */
+  g_free (balsa_app.replyto);
+  if ((field = pl_dict_get_str (globals, "ReplyTo")) == NULL)
+    balsa_app.replyto = g_strdup (balsa_app.address->mailbox);
+  else
+    balsa_app.replyto = g_strdup (field);
 
   /* directory */
   if ((field = pl_dict_get_str (globals, "LocalMailDir")) == NULL)
     return FALSE;
+  g_free (balsa_app.local_mail_directory);
   balsa_app.local_mail_directory = g_strdup (field);
 
+  /* signature file path */
+  g_free (balsa_app.signature_path);
+  if ((field = pl_dict_get_str (globals, "SignaturePath")) == NULL)
+    {
+      balsa_app.signature_path = g_malloc (strlen (g_get_home_dir ()) + 12);
+      sprintf (balsa_app.signature_path, "%s/.signature", g_get_home_dir ());
+    }
+  else
+    balsa_app.signature_path = g_strdup (field);
+
+  if ((field = pl_dict_get_str (globals, "SigSending")) == NULL)
+    balsa_app.sig_sending = TRUE;
+  else
+   balsa_app.sig_sending = atoi ( field );
+
+  if ((field = pl_dict_get_str (globals, "SigReply")) == NULL)
+    balsa_app.sig_whenreply = TRUE;
+  else
+    balsa_app.sig_whenreply = atoi ( field );
+
+  if ((field = pl_dict_get_str (globals, "SigForward")) == NULL)
+    balsa_app.sig_whenforward = TRUE;
+  else
+    balsa_app.sig_whenforward = atoi ( field );
+
   /* smtp server */
   if ((field = pl_dict_get_str (globals, "SMTPServer")) == NULL)
     ;				/* an optional field for now */
+  g_free (balsa_app.smtp_server);
   balsa_app.smtp_server = g_strdup (field);
+ 
+  if ((field = pl_dict_get_str (globals, "SMTP")) == NULL)
+	      balsa_app.smtp = FALSE;
+   else {
+       balsa_app.smtp = atoi (field);
+       if (balsa_app.smtp_server==NULL)
+	       balsa_app.smtp = FALSE;
+   }
+
+  /* Check mail timer */
+  if ((field = pl_dict_get_str (globals, "CheckMailAuto")) == NULL)
+    balsa_app.check_mail_auto = FALSE;
+  else
+    balsa_app.check_mail_auto = atoi (field);
+
+  if ((field = pl_dict_get_str (globals, "CheckMailMinutes")) == NULL)
+    balsa_app.check_mail_timer = 10;
+  else
+    balsa_app.check_mail_timer = atoi (field);
+
+  if (balsa_app.check_mail_timer < 1 )
+    balsa_app.check_mail_timer = 10;
+
+  if( balsa_app.check_mail_auto )
+    update_timer( TRUE, balsa_app.check_mail_timer );
+
+  /* Word Wrap */
+  if ((field = pl_dict_get_str (globals, "WordWrap")) == NULL)
+    balsa_app.wordwrap = TRUE;
+  else
+    balsa_app.wordwrap = atoi (field);
+
+  if ((field = pl_dict_get_str (globals, "WrapLength")) == NULL)
+    balsa_app.wraplength = 79;
+  else
+    balsa_app.wraplength = atoi (field);
+
+  if (balsa_app.wraplength < 40 )
+    balsa_app.wraplength = 40;
 
   /* toolbar style */
   if ((field = pl_dict_get_str (globals, "ToolbarStyle")) == NULL)
@@ -589,12 +720,24 @@ config_global_load (void)
   else
     balsa_app.toolbar_style = atoi (field);
 
-  /* mdi style */
-  if ((field = pl_dict_get_str (globals, "MDIStyle")) == NULL)
-    balsa_app.mdi_style = GNOME_MDI_DEFAULT_MODE;
+  /* Progress Window Dialog */
+  if ((field = pl_dict_get_str (globals, "PWindowOption")) == NULL)
+    balsa_app.pwindow_option = WHILERETR;
   else
-    balsa_app.mdi_style = atoi (field);
+    balsa_app.pwindow_option = atoi (field);
 
+  /* use the preview pane */
+  if ((field = pl_dict_get_str (globals, "UsePreviewPane")) == NULL)
+    balsa_app.previewpane = TRUE;
+  else
+    balsa_app.previewpane = atoi (field);
+#ifdef BALSA_SHOW_INFO
+  /* show mailbox content info */
+  if ((field = pl_dict_get_str (globals, "ShowMailboxContentInfo")) == NULL)
+    balsa_app.mblist_show_mb_content_info = TRUE;
+  else
+    balsa_app.mblist_show_mb_content_info = atoi (field);
+#endif
   /* debugging enabled */
   if ((field = pl_dict_get_str (globals, "Debug")) == NULL)
     balsa_app.debug = FALSE;
@@ -603,22 +746,86 @@ config_global_load (void)
 
   /* window sizes */
   if ((field = pl_dict_get_str (globals, "MainWindowWidth")) == NULL)
-    balsa_app.mw_width = 670;
+    balsa_app.mw_width = 640;
   else
     balsa_app.mw_width = atoi (field);
 
   if ((field = pl_dict_get_str (globals, "MainWindowHeight")) == NULL)
-    balsa_app.mw_height = 435;
+    balsa_app.mw_height = 480;
   else
     balsa_app.mw_height = atoi (field);
 
-  return TRUE;
+  if ((field = pl_dict_get_str (globals, "MailboxListWidth")) == NULL)
+    balsa_app.mblist_width = 100;
+  else
+    balsa_app.mblist_width = atoi (field);
+
+  /* FIXME this can be removed later */
+  /* PKGW: why comment this out? Breaks my Transfer context menu. */
+  if (balsa_app.mblist_width < 100)
+      balsa_app.mblist_width = 170;
+
+  if ((field = pl_dict_get_str (globals, "NotebookHeight")) == NULL)
+    balsa_app.notebook_height = 170;
+  else
+    balsa_app.notebook_height = atoi (field);
+  /* FIXME this can be removed later */
+  /* PKGW see above */
+  if (balsa_app.notebook_height < 100)
+      balsa_app.notebook_height = 200;
+
+
+
+  /* arp --- LeadinStr for "reply to" leadin. */
+  g_free (balsa_app.quote_str);
+  if ((field = pl_dict_get_str (globals, "LeadinStr")) == NULL)
+    balsa_app.quote_str = g_strdup ("> ");
+  else
+    balsa_app.quote_str = g_strdup (field);
+
+  /* font used to display messages */
+  if ((field = pl_dict_get_str (globals, "MessageFont")) == NULL)
+    balsa_app.message_font = g_strdup (DEFAULT_MESSAGE_FONT);
+  else
+    balsa_app.message_font = g_strdup (field);
+
+  /* more here */
+  g_free(balsa_app.charset);
+  if (( field = pl_dict_get_str (globals, "Charset")) == NULL)
+      balsa_app.charset = g_strdup(DEFAULT_CHARSET);
+  else
+      balsa_app.charset = g_strdup(field);
+  mutt_set_charset (balsa_app.charset);
+
+  if (( field = pl_dict_get_str (globals, "EncodingStyle")) == NULL)
+      balsa_app.encoding_style = /*DEFAULT_ENCODING*/ 2;
+  else
+      balsa_app.encoding_style = atoi(field);
+
+  if (( field = pl_dict_get_str (globals, "PrintCommand")) == NULL) 
+      balsa_app.PrintCommand.PrintCommand = g_strdup("a2ps -d -q %s");
+  else 
+      balsa_app.PrintCommand.PrintCommand = g_strdup(field);
+
+  if (( field = pl_dict_get_str (globals, "PrintLinesize")) == NULL)
+      balsa_app.PrintCommand.linesize = DEFAULT_LINESIZE;
+  else
+      balsa_app.PrintCommand.linesize = atoi(field);
+
+  if (( field = pl_dict_get_str (globals, "PrintBreakline")) == NULL )
+      balsa_app.PrintCommand.breakline = FALSE;
+  else
+      balsa_app.PrintCommand.breakline = atoi(field);
+
+ return TRUE;
 }				/* config_global_load */
 
 gint
 config_global_save (void)
 {
   proplist_t globals, temp_str;
+  char tmp[MAX_PROPLIST_KEY_LEN];
+
 
   g_assert (balsa_app.proplist != NULL);
 
@@ -632,10 +839,12 @@ config_global_save (void)
   PLRelease (temp_str);
 
   /* Create a new dictionary of global configurations */
-  if (balsa_app.real_name != NULL)
-    globals = pl_dict_add_str_str (NULL, "RealName", balsa_app.real_name);
-  if (balsa_app.email != NULL)
-    pl_dict_add_str_str (globals, "Email", balsa_app.email);
+  if (balsa_app.address->personal != NULL)
+    globals = pl_dict_add_str_str (NULL, "RealName", balsa_app.address->personal);
+  if (balsa_app.address->mailbox != NULL)
+    pl_dict_add_str_str (globals, "Email", balsa_app.address->mailbox);
+  if (balsa_app.replyto != NULL)
+    pl_dict_add_str_str (globals, "ReplyTo", balsa_app.replyto);
 
   if (balsa_app.local_mail_directory != NULL)
     pl_dict_add_str_str (globals, "LocalMailDir",
@@ -643,22 +852,102 @@ config_global_save (void)
   if (balsa_app.smtp_server != NULL)
     pl_dict_add_str_str (globals, "SMTPServer", balsa_app.smtp_server);
 
+  if (balsa_app.signature_path != NULL)
+    pl_dict_add_str_str (globals, "SignaturePath",
+			 balsa_app.signature_path);
+
+
   {
-    char tmp[32];
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.sig_sending);
+    pl_dict_add_str_str (globals, "SigSending", tmp);
+
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.sig_whenforward);
+    pl_dict_add_str_str (globals, "SigForward", tmp);
+
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.sig_whenreply);
+    pl_dict_add_str_str (globals, "SigReply", tmp);
+
+
     snprintf (tmp, sizeof (tmp), "%d", balsa_app.toolbar_style);
     pl_dict_add_str_str (globals, "ToolbarStyle", tmp);
 
-    snprintf (tmp, sizeof (tmp), "%d", balsa_app.mdi_style);
-    pl_dict_add_str_str (globals, "MDIStyle", tmp);
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.pwindow_option);
+    pl_dict_add_str_str (globals, "PWindowOption", tmp);
 
     snprintf (tmp, sizeof (tmp), "%d", balsa_app.debug);
     pl_dict_add_str_str (globals, "Debug", tmp);
 
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.previewpane);
+    pl_dict_add_str_str (globals, "UsePreviewPane", tmp);
+
+    if (balsa_app.smtp) 
+    {
+      snprintf (tmp, sizeof (tmp), "%d", balsa_app.smtp);
+      pl_dict_add_str_str (globals, "SMTP", tmp);
+    }
+
+#ifdef BALSA_SHOW_INFO
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.mblist_show_mb_content_info);
+    pl_dict_add_str_str (globals, "ShowMailboxContentInfo", tmp);
+#endif
     snprintf (tmp, sizeof (tmp), "%d", balsa_app.mw_width);
     pl_dict_add_str_str (globals, "MainWindowWidth", tmp);
 
     snprintf (tmp, sizeof (tmp), "%d", balsa_app.mw_height);
     pl_dict_add_str_str (globals, "MainWindowHeight", tmp);
+
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.mblist_width);
+    pl_dict_add_str_str (globals, "MailboxListWidth", tmp);
+
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.notebook_height);
+    pl_dict_add_str_str (globals, "NotebookHeight", tmp);
+
+    snprintf (tmp, sizeof (tmp), "%d", balsa_app.encoding_style);
+    pl_dict_add_str_str (globals, "EncodingStyle", tmp);
+  }
+
+  snprintf (tmp, sizeof (tmp), "%d", balsa_app.check_mail_auto );
+  pl_dict_add_str_str (globals, "CheckMailAuto", tmp);
+
+  snprintf (tmp, sizeof (tmp), "%d", balsa_app.check_mail_timer);
+  pl_dict_add_str_str (globals, "CheckMailMinutes", tmp);
+
+  snprintf (tmp, sizeof (tmp), "%d", balsa_app.wordwrap );
+  pl_dict_add_str_str (globals, "WordWrap", tmp);
+
+  snprintf (tmp, sizeof (tmp), "%d", balsa_app.wraplength);
+  pl_dict_add_str_str (globals, "WrapLength", tmp);
+
+  /* arp --- "LeadinStr" into cfg. */
+  if (balsa_app.quote_str != NULL)
+    pl_dict_add_str_str (globals, "LeadinStr", balsa_app.quote_str);
+  else
+    pl_dict_add_str_str (globals, "LeadinStr", "> ");
+
+  /* message font */
+  if (balsa_app.message_font != NULL)
+    pl_dict_add_str_str (globals, "MessageFont", balsa_app.message_font);
+  else
+    pl_dict_add_str_str (globals, "MessageFont", DEFAULT_MESSAGE_FONT);
+
+  /* encoding */
+  if (balsa_app.charset != NULL)
+      pl_dict_add_str_str(globals, "Charset", balsa_app.charset);
+  else
+      pl_dict_add_str_str(globals, "Charset", DEFAULT_CHARSET);
+
+
+  if (balsa_app.PrintCommand.PrintCommand != NULL)
+      pl_dict_add_str_str(globals, "PrintCommand", balsa_app.PrintCommand.PrintCommand);
+  else
+      pl_dict_add_str_str(globals, "PrintCommand", "a2ps -d -q %s");
+
+  {
+      char tmp[MAX_PROPLIST_KEY_LEN];
+      snprintf (tmp, sizeof (tmp), "%d", balsa_app.PrintCommand.linesize);
+      pl_dict_add_str_str (globals, "PrintLinesize", tmp);
+      snprintf (tmp, sizeof (tmp), "%d", balsa_app.PrintCommand.breakline);
+      pl_dict_add_str_str (globals, "PrintBreakline", tmp);
   }
 
   /* Add it to configuration file */
@@ -831,3 +1120,4 @@ config_mailbox_get_highest_number (proplist_t accounts)
 
   return max;
 }				/* config_mailbox_get_highest_number */
+
