@@ -19,18 +19,31 @@
 #ifndef __SAVE_RESTORE_H__
 #define __SAVE_RESTORE_H__
 
+#include "libbalsa.h"
+
 /* FIXME XXX This should be a "configure" option.  Stuart? */
 #define BALSA_CONFIG_FILE ".balsarc"
 
-gint config_load(gchar * user_filename);
-gint config_save(gchar * user_filename);
 
-gint config_mailbox_add (Mailbox * mailbox, char * key_arg);
-gint config_mailbox_delete(gchar * name);
+/* arp --- All proplist keys are assumed to be no large than this.
+ * This is to try and minimize the number of magic numbers used when
+ * key buffers are statically allocated on the stack
+ */
+#ifndef MAX_PROPLIST_KEY_LEN
+/* NB!!! This includes space for the terminating \0! */
+#define MAX_PROPLIST_KEY_LEN    32
+#endif
+
+
+gint config_load (gchar * user_filename);
+gint config_save (gchar * user_filename);
+
+gint config_mailbox_add (Mailbox * mailbox, char *key_arg);
+gint config_mailbox_delete (gchar * name);
 gint config_mailbox_update (Mailbox * mailbox, gchar * old_mbox_name);
-gint config_mailboxes_init(void);
+gint config_mailboxes_init (void);
 
-gint config_global_load(void);
-gint config_global_save(void);
+gint config_global_load (void);
+gint config_global_save (void);
 
 #endif /* __SAVE_RESTORE_H__ */
