@@ -1,3 +1,4 @@
+
 /* Balsa E-Mail Client
  * Copyright (C) 1998 Jay Painter and Stuart Parmenter
  *
@@ -25,19 +26,30 @@ extern "C"
 {
 #endif				/* __cplusplus */
 
+  typedef enum
+    {
+      SEND_NORMAL,
+      SEND_REPLY,
+      SEND_REPLY_ALL,
+      SEND_FORWARD,
+      SEND_CONTINUE
+    }
+  SendType;
+
 
   typedef struct _BalsaSendmsg BalsaSendmsg;
 
   struct _BalsaSendmsg
     {
-      GtkWidget *to, *from, *subject, *cc, *bcc;
-      GtkWidget *text;
       GtkWidget *window;
-      gint type;
-      Message*  orig_message;
-  };
+      GtkWidget *to, *from, *subject, *cc, *bcc, *fcc;
+      GtkWidget *attachments;
+      GtkWidget *text;
+      Message *orig_message;
+      SendType type;
+    };
 
-  void sendmsg_window_new (GtkWidget *, BalsaIndex *, gint);
+  void sendmsg_window_new (GtkWidget *, Message *, SendType);
 
 #ifdef __cplusplus
 }
